/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		arrays.c						 *
* Last rev:								 *
* mods:									 *
* comments:	Array Manipulation Routines	                         *
*									 *
*************************************************************************/

#include "Yap.h"
#include "Yatom.h"
#include "eval.h"
#ifdef TERM_EXTENSIONS
#include "heapgc.h"
#endif
#if HAVE_ERRNO_H
#include <errno.h>
#else
extern int errno;
#endif
#include <string.h>

STATIC_PROTO(Int  p_compile_array_refs, (void));
STATIC_PROTO(Int  p_array_refs_compiled, (void));
STATIC_PROTO(Int  p_sync_mmapped_arrays, (void));

#ifdef YAP_ARRAYS

#ifdef TERM_EXTENSIONS

#include "corout_utils.h"

#endif /* YAP_EXTENSIONS */

#ifdef TERM_EXTENSIONS
STATIC_PROTO(CELL *DestroyNamedArray, (int, CELL *));
STATIC_PROTO(CELL  BindArrayElement, (int, CELL, CELL));
STATIC_PROTO(Term  AccessArrayElement, (Term));
STATIC_PROTO(void  MarkArrayEntry, (CELL *));
STATIC_PROTO(CELL *MarkTrailEntry, (CELL *));
#endif /* TERM_EXTENSIONS */

/*
 * 
 * This file works together with pl/arrays.yap and arrays.h.
 * 
 * YAP now supports a very simple notion of arrays. Arrays may be
 * allocated dynamically or statically:
 * 
 * o anonymous arrays are created during execution and allocated
 * in the heap. They have the lifetime of any other other heap
 * object. Any term can be an argument to a dynamic array.
 * 
 * Dynamic arrays are named as a free variable and are
 * initialised with free variables. 
 * 
 * o named arrays are created during execution but allocated
 * in the code space. They have the lifetime of an heap
 * object. Any term can be an argument to a dynamic array.
 * 
 * Named arrays are named with atoms and are initialised with
 * free variables.
 * 
 * o static arrays are allocated in the heap. Their space is
 * never recovered unless explictly said so by the
 * program. Arguments to these arrays must have fixed size,
 * and can only be atomic (at least for now).
 * 
 * Static arrays can be named through an  atom. They are
 * initialised with [].
 * 
 * Users create arrays by a declaration X array Arity. If X is an atom
 * A, then this it is a static array and A's the array name, otherwise
 * X refers to a dynamic array.
 * 
 * As in C, arrays start counting from 0.
 * 
 * Users access arrays by a token X[I] or a[I], this token can appear
 * anywhere within the computation, so a[2] = X[3*4] means that the
 * second element of global array a should unify with the 12th element
 * of array X. The mechanism used to implement this is the same
 * mechanism used to implement suspension variables.
 * 
 * Representation:
 * 
 * Dynamic Arrays are represented as a compound term of arity N, where
 * N is the size of the array. Even so, I will not include array bound
 * checking for now.
 * 
 * |--------------------------------------------------------------|
 * | $ARRAY/N|....
 * |______________________________________________________________
 * 
 * 
 * Unbound Var is used as a place to point to.
 * 
 * Static Arrays are represented as a special property for an atom,
 * with field size and 
 * 
 * A term of the form X[I] is represented as a Reference pointing to
 * the compound term:
 * 
 * '$array_arg'(X,I)
 * 
 * Dereferecing will automatically find X[I].
 * 
 * The only exception is the compiler, which uses a different
 * dereferencing routine. The clause cl(a[2], Y[X], Y) will be
 * compiled as:
 * 
 * cl(A, B, Y) :- '$access_array'(a, A, 2), '$access_array'(Y, B, X).
 * 
 * There are three operations to access arrays:
 * 
 * X[I] = A, This is normal unification.
 * 
 * X[I] := A, This is multiassignment, and therefore
 * backtrackable.
 * 
 * X[I] ::= A, This is non-backtrackable multiassignment, ans most
 * useful for static arrays.
 * 
 * The LHS of := and of ::= must be an array element!
 * 
 */

Term STATIC_PROTO(AccessNamedArray, (Atom, Int));
void STATIC_PROTO(InitNamedArray, (ArrayEntry *, Int, AtomEntry *));
void STATIC_PROTO(CreateNamedArray, (PropEntry *, Int, Atom));
Int  STATIC_PROTO(p_create_array, (void));
Int  STATIC_PROTO(p_create_mmapped_array, (void));
void STATIC_PROTO(replace_array_references_complex, (CELL *, CELL *, CELL *, Term));
Term STATIC_PROTO(replace_array_references, (Term));
Int  STATIC_PROTO(p_array_references, (void));
Int  STATIC_PROTO(p_create_static_array, (void));
Int  STATIC_PROTO(p_access_array, (void));
Int  STATIC_PROTO(p_assign_static, (void));

static Term
AccessNamedArray(Atom a, Int indx)
{
  ArrayEntry *pp = RepArrayProp(RepAtom(a)->PropOfAE);

  while (!EndOfPAEntr(pp) && pp->KindOfPE != ArrayProperty)
    pp = RepArrayProp(pp->NextOfPE);

  if (!EndOfPAEntr(pp)) {
    if (ArrayIsDynamic(pp)) {
      return (((CELL *)*(pp->ValueOfVE))[indx]);
    } else {
      StaticArrayEntry *ptr = (StaticArrayEntry *)pp;
      if (-(pp->ArrayEArity) <= indx || indx < 0) {
	/*	Error(DOMAIN_ERROR_ARRAY_OVERFLOW, MkIntegerTerm(indx), "access_array");*/
	P = (yamop *)FAILCODE;
	return(TermNil);
      }
      switch (ptr->ArrayType) {
      case array_of_ints:
	return (MkIntegerTerm((ptr->ValueOfVE[indx])));
      case array_of_doubles:
	return (MkEvalFl(*(Float *)(ptr->ValueOfVE+indx*RSDbl)));
      case array_of_ptrs:
	return (MkGeneralRefTerm((AtomEntry *)(ptr->ValueOfVE[indx])));
      case array_of_atoms:
	/* just return the atom */
	return (ptr->ValueOfVE[indx]);
      case array_of_dbrefs:
	{
	  /* The object is now in use */
	  Term TRef = ptr->ValueOfVE[indx];

	  if (TRef != TermNil) {
	    DBRef ref = DBRefOfTerm(TRef);
	    if (!(ref->Flags & InUseMask)) {
	      ref->Flags |= InUseMask;
	      TRAIL_REF(&(ref->Flags));	/* So that fail will erase it */
	    }
	  } else {
	    P = (yamop *)FAILCODE;
	  }
	  return (TRef);
	}
      default:
	return(TermNil);
      }
    }      
  }
  else {
    Error(EXISTENCE_ERROR_ARRAY,MkAtomTerm(a),"named array");
    return (TermNil);
  }

}

#ifdef TERM_EXTENSIONS

/* Routine to access an array element */
static Term 
AccessArrayElement(Term inp)
{
  register Int indx;
  array_access *p;
  register Term t;
  Term ti;

  p = (array_access *)RepAppl(inp);
  t = Deref(p->ArrayT);
  ti = Deref(p->indx);
  if (IsNonVarTerm(ti)) {
    if (IsIntTerm(ti))
      indx = IntOfTerm(ti);
    else if (Eval(ti) == FInt)
      indx = reg.IntMode;
    else {
      Error(TYPE_ERROR_INTEGER,ti,"array index");
      return (TermNil);
    }
  }
  else {
    Error(INSTANTIATION_ERROR,ti,"array index");
    return (TermNil);
  }

  if (IsNonVarTerm(t)) {
    if (IsApplTerm(t)) {
      if (indx >= ArityOfFunctor(FunctorOfTerm(t)) || indx < 0) {
	/*	Error(DOMAIN_ERROR_ARRAY_OVERFLOW, MkIntegerTerm(indx), "access_array");*/
	P = (yamop *)FAILCODE;
	return(TermNil);
      }
      return (Derefa((RepAppl(t))+)indx + 1)));
    }
    else if (IsAtomTerm(t)) {
      return (Deref(AccessNamedArray(AtomOfTerm(t), indx)));
    }
    else
      Error(TYPE_ERROR_ARRAY,t,"access array element");
  }
  else
    Error(INSTANTIATION_ERROR,t,"access array element");
  return (TermNil);
}

static void
MarkArrayEntry(CELL *p)
{
  array_access *ap = (array_access *)p;

  if (MARKED(*p))
    return;
  MARK(p);
  total_marked += 3;
  MARK(((CELL *)&(ap->ArrayT)));
  MARK(((CELL *)&(ap->indx)));
}

static CELL *
MarkTrailEntry(CELL *trailp)
{
  /* given a key to an array, just take it off-line */
  PropEntry *pp = RepProp(((AtomEntry *)RepAppl(trailp[-1]))->PropOfAE);
  
  while (!EndOfPAEntr(pp) && pp->KindOfPE != ArrayProperty) {
    pp = RepProp(pp->NextOfPE);
  }
  if (!EndOfPAEntr(pp)) {
    ArrayEntry * p = ((ArrayEntry *) pp);
    Int dim = p->ArrayEArity;
    CELL *ptr =  (CELL *)*(p->ValueOfVE);

    if (HEAP_PTR(trailp[-2])) {
      mark_variable(trailp-2);
      total_marked--;
    }
    for (; dim > 0; dim--) {
     mark_variable(ptr++);
    }
    return(trailp-3);
  } else
    return(trailp-1);
}

#endif /* TERM_EXTENSIONS */

static Int 
p_access_array(void)
{
  Term t = Deref(ARG1);
  Term ti = Deref(ARG2);
  Term tf;
  Int indx;

  if (IsNonVarTerm(ti)) {
    if (IsIntTerm(ti))
      indx = IntOfTerm(ti);
    else if (Eval(ti) == FInt)
      indx = reg.IntMode;
    else {
      Error(TYPE_ERROR_INTEGER,ti,"access_array");
      return (FALSE);
    }
  }
  else {
    Error(INSTANTIATION_ERROR,ti,"access_array");
    return (TermNil);
  }

  if (IsNonVarTerm(t)) {
    if (IsApplTerm(t)) {
      if (indx >= ArityOfFunctor(FunctorOfTerm(t)) || indx < 0) {
	/*	Error(DOMAIN_ERROR_ARRAY_OVERFLOW, MkIntegerTerm(indx), "access_array");*/
	P = (yamop *)FAILCODE;
	return(FALSE);
      }
      tf = Derefa((RepAppl(t))+(indx + 1));
    } else if (IsAtomTerm(t)) {
      tf = AccessNamedArray(AtomOfTerm(t), indx);
    } else {
      Error(TYPE_ERROR_ARRAY,t,"access_array");
      return(FALSE);
    }    
  } else {
    Error(INSTANTIATION_ERROR,t,"access_array");
    return(FALSE);
  }
  return (unify(tf, ARG3));
}

static Int 
p_array_arg(void)
{
  register Term ti = Deref(ARG3), t;
  register Int indx;

  if (IsNonVarTerm(ti)) {
    if (IsIntTerm(ti))
      indx = IntOfTerm(ti);
    else if (Eval(ti) == FInt)
      indx = reg.IntMode;
    else {
      Error(TYPE_ERROR_INTEGER,ti,"array_arg");
      return (FALSE);
    }
  }
  else {
    Error(INSTANTIATION_ERROR,ti,"array_arg");
    return (FALSE);
  }

  t = Deref(ARG2);
  if (IsNonVarTerm(t)) {
    if (IsApplTerm(t)) {
      return (unify(((RepAppl(t))[indx + 1]), ARG1));
    }
    else if (IsAtomTerm(t)) {
      Term tf = AccessNamedArray(AtomOfTerm(t), indx);

      return (unify(tf, ARG1));
    }
    else
      Error(TYPE_ERROR_ARRAY,t,"array_arg");
  }
  else
    Error(INSTANTIATION_ERROR,t,"array_arg");

  return (FALSE);

}

#ifdef TERM_EXTENSIONS

/* Routine to dynamically bind an array element and then continue
 * execution. Called from unification and from the abstract machine
 * emulator */
static CELL 
BindArrayElement(int bind_op, CELL reg1, CELL reg2)
{
  Term t = AccessArrayElement(reg1);
  
  switch (bind_op) {
  case SBIND:
    return (unify(reg2, t));
  case SISPAIR:
  case SISAPPL:
    return (t);
  }
  return (0);
}

#endif /* TERM_EXTENSIONS */

static void
InitNamedArray(ArrayEntry * p, Int dim, AtomEntry *ap)
{
  Term *tp;

  /* Leave a pointer so that we can reclaim array space when
   * we backtrack or when we abort */
  /* place terms in reverse order */
  p->ValueOfVE = TR;
  tp = H;
  TR[0] = (CELL)H;
  TR[1] = AbsAppl(((CELL *)ap));
  TR[2] = AbsAppl((CELL *)(FunctorArrayAccess));
  TR += 3;
  p->ArrayEArity = dim;
  /* Initialise the array as a set of variables */
  H += dim;
  for (; tp < H; tp++) {
    *tp = (Term) tp;
  }
}

static void
CreateNamedArray(PropEntry * pp, Int dim, Atom ap)
{
  ArrayEntry *p;

  YAPEnterCriticalSection();
  p = (ArrayEntry *) AllocAtomSpace(sizeof(*p));
  p->KindOfPE = ArrayProperty;
  p->NextOfPE = RepAtom(ap)->PropOfAE;
  RepAtom(ap)->PropOfAE = AbsArrayProp(p);
  YAPLeaveCriticalSection();

  InitNamedArray(p, dim, RepAtom(ap));

}

static void
CreateStaticArray(Atom at, Int dim, static_array_types type, CODEADDR start_addr)
{
  StaticArrayEntry *p;
  int i, array_size;

  YAPEnterCriticalSection();
  if (type == array_of_doubles)
    array_size = dim*(RSDbl);
  else
    array_size = dim;
  p = (StaticArrayEntry *) AllocAtomSpace(sizeof(*p));
  p->KindOfPE = ArrayProperty;
  p->NextOfPE = RepAtom(at)->PropOfAE;
  p->ArrayEArity = -dim;
  p->ArrayType = type;
  if (start_addr == NULL) {
    while ((p->ValueOfVE = (CELL *) AllocAtomSpace(array_size*sizeof(CELL))) == NULL) {
      YAPLeaveCriticalSection();
      if (!growheap()) {
	Abort("[ SYSTEM ERROR: YAP failed to reserve space in growheap ]\n");
	return;
      }
      YAPEnterCriticalSection();
    }
    switch(type) {
    case array_of_ints:
      for (i = 0; i < dim; i++)
	p->ValueOfVE[i] = 0;
      break;
    case array_of_doubles:
      {
	union {
	  Float f;
	  CELL ar[RSDbl];
	} un;
	int j;

	/* avoid alignment trouble */
	un.f = 0.0;
	for (i = 0; i < dim; i++)
	  for (j = 0; j < RSDbl; j++)
	    p->ValueOfVE[i*RSDbl+j] = un.ar[j];
	break;
      }
    case array_of_ptrs:
      {
	AtomEntry **ptr = (AtomEntry **)(p->ValueOfVE);
	for (i = 0; i < dim; i++)
	  *ptr++ = NULL;
	break;
      }
    case array_of_atoms:
      {
	Term *ptr = p->ValueOfVE;
	for (i = 0; i < dim; i++)
	  *ptr++ = TermNil;
	break;
      }
    case array_of_dbrefs:
      {
	CELL *ptr = p->ValueOfVE;
	for (i = 0; i < dim; i++)
	  *ptr++ = TermNil;
	break;
      }
    }
  } else {
    p->ValueOfVE = (CELL *) start_addr;
  }
  RepAtom(at)->PropOfAE = AbsArrayProp((ArrayEntry *)p);
  YAPLeaveCriticalSection();
}

CELL * 
ClearNamedArray(CELL *pt0)
{
  /* given a key to an array, just take it off-line */
  PropEntry *pp = RepProp(((AtomEntry *)RepAppl(pt0[-1]))->PropOfAE);
  
  while (!EndOfPAEntr(pp) && pp->KindOfPE != ArrayProperty) {
    pp = RepProp(pp->NextOfPE);
  }
  if (!EndOfPAEntr(pp)) {
    ((ArrayEntry *) pp)->ArrayEArity = 0;
    /* tell backtracking to skip two cells */
    return(pt0-2);
  } else {
    Error(EXISTENCE_ERROR_ARRAY,TermNil,"clear array");
    return(pt0); /* just make GCC happy */
  }
}

#ifdef TERM_EXTENSIONS
static CELL * 
DestroyNamedArray(int caller, CELL *pt0)
{
  switch (caller) {
  case OP_UNBIND:
    return(ClearNamedArray(pt0));
  default:
    /* just do nothing! */
    return(pt0);
  }
}
#endif /* TERM_EXTENSIONS */

/* create an array (?Name, + Size) */
static Int 
p_create_array()
{
  Term ti = Deref(ARG2);
  Term t = Deref(ARG1);
  Int size;

  if (IsIntTerm(ti))
    size = IntOfTerm(ti);
  else if (Eval(ti) == FInt) {
    size = reg.IntMode;
  }
  else {
    Error(TYPE_ERROR_INTEGER,ti,"create_array");
    return (FALSE);
  }

  if (IsVarTerm(t)) {
    /* Create an anonymous array */
    Functor farray;

    farray = MkFunctor(AtomArray, size);
    t = AbsAppl(H);
    *H++ = (CELL) farray;
    for (; size >= 0; size--) {
      RESET_VARIABLE(H);
      H++;
    }
    return (unify(t, ARG1));
  }
  else if (IsAtomTerm(t)) {
    /* Create a named array */
    Atom a = AtomOfTerm(t);
    PropEntry *pp = RepProp(RepAtom(a)->PropOfAE);

    while (!EndOfPAEntr(pp) && pp->KindOfPE != ArrayProperty)
      pp = RepProp(pp->NextOfPE);
    if (EndOfPAEntr(pp)) {
      CreateNamedArray(pp, size, a);
      return (TRUE);
    }
    else {
      ArrayEntry *app = (ArrayEntry *) pp;

      if (app->ArrayEArity)
	Error(PERMISSION_ERROR_CREATE_ARRAY,t,"create_array",
	      RepAtom(a)->StrOfAE);
      else {
	InitNamedArray(app, size, RepAtom(a));
	return (TRUE);
      }
    }
  }
  return (FALSE);
}

/* create an array (+Name, + Size, +Props) */
static Int 
p_create_static_array()
{
  Term ti = Deref(ARG2);
  Term t = Deref(ARG1);
  Term tprops = Deref(ARG3);
  Int size;
  static_array_types props;

  if (IsVarTerm(ti)) {
    Error(INSTANTIATION_ERROR,ti,"create static array");
    return (FALSE);
  } else if (IsIntTerm(ti))
    size = IntOfTerm(ti);
  else if (Eval(ti) == FInt) {
    size = reg.IntMode;
  }
  else {
    Error(TYPE_ERROR_INTEGER,ti,"create static array");
    return (FALSE);
  }

  if (IsVarTerm(tprops)) {
    Error(INSTANTIATION_ERROR,tprops,"create static array");
    return (FALSE);
  } else if (IsAtomTerm(tprops)) {
    char *atname = RepAtom(AtomOfTerm(tprops))->StrOfAE;
    if (!strcmp(atname, "int"))
      props = array_of_ints;
    else if (!strcmp(atname, "dbref"))
      props = array_of_dbrefs;
    else if (!strcmp(atname, "float"))
      props = array_of_doubles;
    else if (!strcmp(atname, "ptr"))
      props = array_of_ptrs;
    else if (!strcmp(atname, "atom"))
      props = array_of_atoms;
    else {
      Error(DOMAIN_ERROR_ARRAY_TYPE,tprops,"create static array");
      return(FALSE);
    }
  } else {
    Error(TYPE_ERROR_ATOM,tprops,"create static array");
    return (FALSE);
  }

  if (IsVarTerm(t)) {
    Error(INSTANTIATION_ERROR,t,"create static array");
    return (FALSE);
  }
  else if (IsAtomTerm(t)) {
    /* Create a named array */
    Atom a = AtomOfTerm(t);
    PropEntry *pp = RepProp(RepAtom(a)->PropOfAE);

    while (!EndOfPAEntr(pp) && pp->KindOfPE != ArrayProperty)
      pp = RepProp(pp->NextOfPE);
    if (EndOfPAEntr(pp)) {
      CreateStaticArray(a, size, props, NULL);
      return (TRUE);
    } else {
      Error(PERMISSION_ERROR_CREATE_ARRAY,t,"create static array");
      return(FALSE);
    }
  } else {
    Error(TYPE_ERROR_ATOM,t,"create static array");
    return (FALSE);
  }
}

#if HAVE_MMAP

#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>

/* keep a list of mmaped blocks to synch on exit */

typedef struct MMAP_ARRAY_BLOCK {
  Atom name;
  void *start;
  size_t  uses;
  struct MMAP_ARRAY_BLOCK *next;
} mmap_array_block;

static mmap_array_block *mmap_arrays = NULL;

#endif

/* create an array (+Name, + Size, +Props) */
static Int 
p_create_mmapped_array()
{
#ifdef HAVE_MMAP
  Term ti = Deref(ARG2);
  Term t = Deref(ARG1);
  Term tprops = Deref(ARG3);
  Term tfile = Deref(ARG4);
  Int size;
  static_array_types props;
  size_t total_size;
  CODEADDR array_addr;

  if (IsVarTerm(ti)) {
    Error(INSTANTIATION_ERROR,ti,"create_mmapped_array");
    return (FALSE);
  } else if (IsIntTerm(ti))
    size = IntOfTerm(ti);
  else if (Eval(ti) == FInt) {
    size = reg.IntMode;
  }
  else {
    Error(TYPE_ERROR_INTEGER,ti,"create_mmapped_array");
    return (FALSE);
  }

  if (IsVarTerm(tprops)) {
    Error(INSTANTIATION_ERROR,tprops,"create_mmapped_array");
    return (FALSE);
  } else if (IsAtomTerm(tprops)) {
    char *atname = RepAtom(AtomOfTerm(tprops))->StrOfAE;
    if (!strcmp(atname, "int")) {
      props = array_of_ints;
      total_size = size*sizeof(Int);
    } else if (!strcmp(atname, "dbref")) {
      props = array_of_dbrefs;
      total_size = size*sizeof(Int);
    } else if (!strcmp(atname, "float")) {
      props = array_of_doubles;
      total_size = size*sizeof(Float);
    } else if (!strcmp(atname, "ptr")) {
      props = array_of_ptrs;
      total_size = size*sizeof(AtomEntry *);
    } else if (!strcmp(atname, "atom")) {
      props = array_of_atoms;
      total_size = size*sizeof(Term);
    } else {
      Error(DOMAIN_ERROR_ARRAY_TYPE,tprops,"create_mmapped_array");
      return(FALSE);
    }
  } else {
    Error(TYPE_ERROR_ATOM,tprops,"create_mmapped_array");
    return (FALSE);
  }

  if (IsVarTerm(tfile)) {
    Error(INSTANTIATION_ERROR,tfile,"create_mmapped_array");
    return (FALSE);
  } else if (IsAtomTerm(tfile)) {
    char *filename = RepAtom(AtomOfTerm(tfile))->StrOfAE;
    int fd = open(filename, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR);

    if (lseek(fd, total_size-1, SEEK_SET) < 0)
      Error(SYSTEM_ERROR,tfile,"create_mmapped_array (lseek: %s)", strerror(errno));
    if (write(fd, "", 1) < 0)
      Error(SYSTEM_ERROR,tfile,"create_mmapped_array (write: %s)", strerror(errno));
    /*
      if (ftruncate(fd, total_size) < 0)
      Error(SYSTEM_ERROR,tfile,"create_mmapped_array");
    */
    if ((array_addr = (CODEADDR)mmap(0, (size_t) total_size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0)) == (CODEADDR) - 1)
      Error(SYSTEM_ERROR,tfile,"create_mmapped_array (mmap: %s)", strerror(errno));
    if (close(fd) < 0)
      Error(SYSTEM_ERROR,tfile,"create_mmapped_array (close: %s)", strerror(errno));

  } else {
    Error(TYPE_ERROR_ATOM,tfile,"create_mmapped_array");
    return (FALSE);
  }

  if (IsVarTerm(t)) {
    Error(INSTANTIATION_ERROR,t,"create_mmapped_array");
    return (FALSE);
  }
  else if (IsAtomTerm(t)) {
    /* Create a named array */
    Atom a = AtomOfTerm(t);
    PropEntry *pp = RepProp(RepAtom(a)->PropOfAE);

    while (!EndOfPAEntr(pp) && pp->KindOfPE != ArrayProperty)
      pp = RepProp(pp->NextOfPE);
    if (EndOfPAEntr(pp)) {
      mmap_array_block *ptr;

      CreateStaticArray(a, size, props, array_addr);
      ptr = (mmap_array_block *)AllocAtomSpace(sizeof(mmap_array_block));
      ptr->name = a;
      ptr->uses = total_size;
      ptr->start = (void *)array_addr;
      ptr->next = mmap_arrays;
      mmap_arrays = ptr;
      return(TRUE);
    } else {
      Error(DOMAIN_ERROR_ARRAY_TYPE,t,"create_mmapped_array", RepAtom(a)->StrOfAE);
      return(FALSE);
    }
  } else {
    Error(TYPE_ERROR_ATOM,t,"create_mmapped_array");
    return (FALSE);
  }
#else
  Error(SYSTEM_ERROR,TermNil,"create_mmapped_array (mmap)");
  return (FALSE);
#endif
}

#ifdef RATIONAL_TREES

#define UNWIND_CUNIF()                                        \
         while (visited < AuxSp) {                            \
            pt0 = (CELL *)visited[0];                         \
            *pt0 = visited[1];                                \
            visited += 2;                                     \
         }

#else
#define UNWIND_CUNIF()
#endif

/* This routine verifies whether a complex has variables. */
static void 
replace_array_references_complex(register CELL *pt0,
				 register CELL *pt0_end,
				 register CELL *ptn,
				 Term Var)
{

#ifdef RATIONAL_TREES
  register CELL *visited = AuxSp;

#endif
  register CELL **to_visit = (CELL **) PreAllocCodeSpace();
  CELL **to_visit_base = to_visit;

loop:
  while (pt0 < pt0_end) {
    register CELL d0;

    ++pt0;
#ifdef TERM_EXTENSIONS
    d0 = SDerefa(pt0);
#else
    d0 = Derefa(pt0);
#endif /* TERM_EXTENSIONS */
  restart:
    if (IsVarTerm(d0)) {
      *ptn++ = d0;
    }
    else if (IsPairTerm(d0)) {
#ifdef RATIONAL_TREES
      /* now link the two structures so that no one else will */
      /* come here */
      visited -= 2;
      visited[0] = (CELL) pt0;
      visited[1] = *pt0;
      *pt0 = TermNil;
#endif
      /* store the terms to visit */
      *ptn++ = AbsPair(H);
      if (pt0 < pt0_end) {
	to_visit[0] = pt0;
	to_visit[1] = pt0_end;
	to_visit[2] = ptn;
	to_visit += 3;
      }
      pt0 = RepPair(d0) - 1;
      pt0_end = RepPair(d0) + 1;
      /* write the head and tail of the list */
      ptn = H;
      H += 2;
    }
    else if (IsApplTerm(d0)) {
      register Functor f;

      f = FunctorOfTerm(d0);
      /* store the terms to visit */
      if (IsExtensionFunctor(f)) {
	if (IsAttachFunc(f)) {
	  if (f == FunctorArrayAccess) {
	    Term tar[3], tah, tat, nvar;

	    array_access *acc = (array_access *) (RepAppl(d0)+1);
	    tar[0] = tah = *ptn++ = MkVarTerm();
	    tar[1] = acc->ArrayT;
	    tar[2] = acc->indx;
	    tat = MkApplTerm(FunctorArrayEntry, 3, tar);
	    nvar = MkVarTerm();
	    Bind(PtrOfTerm(Var), MkPairTerm(tat, nvar));
	    Var = nvar;
	  } else {
	    /* obtain the term */
	    d0 = attas[ExtFromFunctor(f)].deref_op(d0);
	    /* ignore reference chains */
	    goto restart;
	  } 
	} else {
	  *ptn++ = d0;
	  continue;
	}
      }
#ifdef RATIONAL_TREES
      /* now link the two structures so that no one else will */
      /* come here */
      visited -= 2;
      visited[0] = (CELL) pt0;
      visited[1] = *pt0;
      *pt0 = TermNil;
#endif
      *ptn++ = AbsAppl(H);
      /* store the terms to visit */
      if (pt0 < pt0_end) {
	to_visit[0] = pt0;
	to_visit[1] = pt0_end;
	to_visit[2] = ptn;
	to_visit += 3;
      }
      pt0 = RepAppl(d0);
      d0 = ArityOfFunctor(f);
      pt0_end = pt0 + d0;
      /* start writing the compound term */
      ptn = H;
      *ptn++ = (CELL) f;
      H += d0 + 1;
    }
    else {			/* AtomOrInt */
      *ptn++ = d0;
    }
    /* just continue the loop */
  }

  /* Do we still have compound terms to visit */
  if (to_visit > (CELL **) to_visit_base) {
    to_visit -= 3;
    pt0 = to_visit[0];
    pt0_end = to_visit[1];
    ptn = to_visit[2];
    goto loop;
  }

  UNWIND_CUNIF();
  Bind(PtrOfTerm(Var), TermNil);

}

/*
 * 
 * Given a term t0, build a new term tf of the form ta+tb, where ta is
 * obtained by replacing the array references in t0 by empty
 * variables, and tb is a list of array references and corresponding
 * variables.
 */
static Term 
replace_array_references(Term t0)
{
  Term t;

#ifdef TERM_EXTENSIONS
  t = SDeref(t0);
#else
  t = Deref(t0);
#endif /* TERM_EXTENSIONS */
  do {
  restart_replace_array_defs:
    if (IsVarTerm(t)) {
      /* we found a variable */
      return (MkPairTerm(t, TermNil));
    }
    else if (IsAtomOrIntTerm(t)) {
      return (MkPairTerm(t, TermNil));
    }
    else if (IsPairTerm(t)) {
      Term VList = MkVarTerm();
      CELL *h0 = H;

      H += 2;
      replace_array_references_complex(RepPair(t) - 1, RepPair(t) + 1, h0,
				       VList);
      return (MkPairTerm(AbsPair(h0), VList));
    }
    else {
      Term VList = MkVarTerm();
      CELL *h0 = H;
      Functor f = FunctorOfTerm(t);

      if (IsAttachFunc(f)) {
	if (f == FunctorArrayAccess) {
	  Term tar[3], tah, tat;

	  array_access *acc = (array_access *) (RepAppl(t)+1);

	  tar[0] = tah = MkVarTerm();
	  tar[1] = acc->ArrayT;
	  tar[2] = acc->indx;
	  tat = MkApplTerm(FunctorArrayEntry, 3, tar);
	  return (MkPairTerm(tah, MkPairTerm(tat, TermNil)));
	}
	else {
	  /* obtain the term */
	  t = attas[ExtFromFunctor(f)].deref_op(t);
	  /* ignore reference chains */
	  goto restart_replace_array_defs;
	}
      }
      *H++ = (CELL) (f);
      H += ArityOfFunctor(f);
      replace_array_references_complex(RepAppl(t),
		      RepAppl(t) + ArityOfFunctor(FunctorOfTerm(t)), h0 + 1,
				       VList);
      return (MkPairTerm(AbsAppl(h0), VList));
    }
  } while (TRUE);
  /* make lcc happy */
  return(FALSE);
}

static Int 
p_array_references(void)
{
  Term t = replace_array_references(ARG1);
  Term t1 = HeadOfTerm(t);
  Term t2 = TailOfTerm(t);

  return (unify(t1, ARG2) && unify(t2, ARG3));
}

static Int 
p_assign_static(void)
{
  Term t1, t2, t3;
  StaticArrayEntry *ptr;
  int indx;

  t1 = Deref(ARG1);
  if (IsVarTerm(t1)) {
    Error(INSTANTIATION_ERROR,t1,"assign_static");
    return(FALSE);
  }
  if (!IsAtomTerm(t1)) {
    Error(TYPE_ERROR_ATOM,t1,"assign_static");
    return(FALSE);
  }
  ptr =  (StaticArrayEntry *)RepArrayProp(RepAtom(AtomOfTerm(t1))->PropOfAE);

  while (!EndOfPAEntr(ptr) && ptr->KindOfPE != ArrayProperty)
    ptr = (StaticArrayEntry *)RepArrayProp(ptr->NextOfPE);

  if (EndOfPAEntr(ptr) ||
      ArrayIsDynamic((ArrayEntry *)ptr)) {
    Error(EXISTENCE_ERROR_ARRAY,t1,"%s in assign_static", RepAtom(AtomOfTerm(t1))->StrOfAE);
    return(FALSE);
  }

  t2 = Deref(ARG2);
  if (IsNonVarTerm(t2)) {
    if (IsIntTerm(t2))
      indx = IntOfTerm(t2);
    else if (Eval(t2) == FInt)
      indx = reg.IntMode;
    else {
      Error(TYPE_ERROR_INTEGER,t2,"assign_static");
      return (FALSE);
    }
  } else {
    Error(INSTANTIATION_ERROR,t2,"assign_static");
    return (FALSE);
  }
    

  t3 = Deref(ARG3);
  if (IsVarTerm(t3)) {
    Error(INSTANTIATION_ERROR,t3,"assign_static");
    return (FALSE);
  }
    

  switch (ptr->ArrayType) {
  case array_of_ints:
    {
      Int i;
      
      if (IsIntTerm(t3))
	i = IntOfTerm(t3);
      else if (Eval(t3) == FInt)
	i = reg.IntMode;
      else {
	Error(TYPE_ERROR_INTEGER,t3,"assign_static");
	return (FALSE);
      }
      ptr->ValueOfVE[indx]= i;
    }
    break;

  case array_of_doubles:
    {
      CELL *p = ptr->ValueOfVE + indx*RSDbl;
      union {
	Float f;
	CELL ar[RSDbl];
      } un;
      int i;

      if (IsFloatTerm(t3))
	un.f = FloatOfTerm(t3);
      else if (Eval(t3) == FFloat)
	un.f = reg.FloatMode;
      else {
	Error(TYPE_ERROR_FLOAT,t3,"assign_static");
	return (FALSE);
      }
      for (i = 0; i < RSDbl; i++)
	p[i] = un.ar[i];

    }
    break;

  case array_of_ptrs:
    {
      CODEADDR r;

      if (IsGeneralRefTerm(t3))
	r = GeneralRefOfTerm(t3);
      else {
	Error(TYPE_ERROR_PTR,t3,"assign_static");
	return (FALSE);
      }
      ptr->ValueOfVE[indx]= (CELL)r;
    }
    break;

  case array_of_atoms:
    {
      if (!IsAtomTerm(t3)) {
	Error(TYPE_ERROR_ATOM,t3,"assign_static");
	return (FALSE);
      }
      ptr->ValueOfVE[indx]= t3;
    }
    break;

  case array_of_dbrefs:
    {
      
      Term t0 = ptr->ValueOfVE[indx];
      
      if (!IsDBRefTerm(t3)) {
	Error(TYPE_ERROR_DBREF,t3,"assign_static");
	return (FALSE);
      }
      ptr->ValueOfVE[indx]= t3;
      if (t0 != TermNil)
	DBRefOfTerm(t0)->NOfRefsTo--;
      DBRefOfTerm(t3)->NOfRefsTo++;
    }
    break;
  }
  return(TRUE);
}
#endif

#ifdef YAP_ARRAYS
int compile_arrays = FALSE;

#endif

static Int 
p_compile_array_refs(void)
{
#ifdef YAP_ARRAYS
  compile_arrays = TRUE;
#endif
  return (TRUE);
}

static Int 
p_array_refs_compiled(void)
{
#ifdef YAP_ARRAYS
  return (compile_arrays);
#else
  return (FALSE);
#endif
}

static Int
p_sync_mmapped_arrays()
{
#ifdef YAP_ARRAYS
#ifdef HAVE_MMAP
  mmap_array_block *ptr = mmap_arrays;
  while (ptr != NULL) {
    msync(ptr->start, ptr->uses, MS_SYNC);
    ptr = ptr->next;
  }
#endif
#endif
  return(TRUE);
}

void 
InitArrayPreds(void)
{
#ifdef YAP_ARRAYS
#ifdef TERM_EXTENSIONS
  attas[array_acc_ext].bind_op = BindArrayElement;
  attas[array_acc_ext].deref_op = AccessArrayElement;
  attas[array_acc_ext].untrail_op = DestroyNamedArray;
  attas[array_acc_ext].binding_transparent = FALSE;
  attas[array_acc_ext].mark_op = MarkArrayEntry;
  attas[array_acc_ext].mark_from_trail_op = MarkTrailEntry;
#endif /* TERM_EXTENSIONS */
  InitCPred("$create_array", 2, p_create_array, SafePredFlag|SyncPredFlag);
  InitCPred("$array_references", 3, p_array_references, SafePredFlag);
  InitCPred("$array_arg", 3, p_array_arg, SafePredFlag);
  InitCPred("static_array", 3, p_create_static_array, SafePredFlag|SyncPredFlag);
  InitCPred("mmapped_array", 4, p_create_mmapped_array, SafePredFlag|SyncPredFlag);
  InitCPred("update_array", 3, p_assign_static, SafePredFlag);
  InitCPred("array_element", 3, p_access_array, SafePredFlag);
#endif
  InitCPred("$sync_mmapped_arrays", 0, p_sync_mmapped_arrays, SafePredFlag);
  InitCPred("$compile_array_refs", 0, p_compile_array_refs, SafePredFlag);
  InitCPred("$array_refs_compiled", 0, p_array_refs_compiled, SafePredFlag);
}

