/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		corout.c						 *
* Last rev:								 *
* mods:									 *
* comments:	Co-routining from within YAP				 *
*									 *
*************************************************************************/
#ifdef SCCS
static char SccsId[]="%W% %G%";
#endif

#include "Yap.h"
#include "Yatom.h"
#include "heapgc.h"

/*

These are simple routines to support co-routining in YAP. The idea is
to make the interface as simple as possible.

The interface for co-routines is:

$freeze(+X,+G) -> execute G only when V is *bound* (not
necessarily ground.

The data-structures are:


                       	|-------------------|
Ref    ---------------->|  V   |   SG\      |
			|-------------\-----|
				       \
					\ |------------/
					 >| |  | G | NS|
					  | |  |   |   |
					  |-|----------|
					    |
					    V
                                          |------------/
					->|    | G | NS|
					| | |  |   |   |
					| |-|----------|
					|   |
					|----

Where V is an indicator for the term, SG is a pointer for the list of
suspended goals, G is the suspended goal, and NS is a pointer to a
list of suspended goals.

When suspend_on is called, it executes the following operations:
	if (X is a reference):
		add a record containing G to the tail
		of the current SG list for V
	if (X is unbound)
		create a record R containing G and a self-reference.
		create a suspension register containing a free
		variable V and a pointer to R (sus goal list)
		Bind V to a Ref to the new structure.
	if (X is nonvar)
		Oooppssss!!!! The Prolog interface should have
		prevented this.

When trying to unify a nonvar to a suspension variable, the following
actions are taken:

	o Bind V to to the nonvar. This is done within absmi.c and
	depends a lot on the surrounding code.

	o Make the list SG the head of the list WokenGoals.

	o Activate the Interrupt Flag, so that the system will process
	the suspended goals at the next "call" absmiop.

	At the next "call":
	 + Save the current goal on the heap (C).
	 + Take the first member of the WokenGoals list.
	 + set up '$resume_and_continue'(?G,?C), which should execute
	   G and then C.
	 + If WokenGoals is empty, down Interrupt Flag
	 + jump to the code for '$resume_and_continue'(?G,?C)
	 + Note, the system will fetch the next goal at the next
	 "call" op.

When trying to unify two suspended variables X and Y, we just bind X
to Y, and include Y's goals in X's list.

The standard definition for resume_and_continue:

'$wake_up_goal'(C,G) :- call(G), call(C).

Advantages:

	o Implementation is simple (the main work is changing absmi).

	o Does not need updatable variables.

	o No special support in backtracking.

	o Data structures spend little space.

Disadvantages

	o We create a goal frame for every suspended goal. This is
	avoided by storing both P and the arguments in the suspension
	record, and then jumping.

	o We do a lot of meta-calls. This can be avoided by
	manipulating P and CP directly.

*/

STATIC_PROTO(Int p_read_svar_list, (void));
STATIC_PROTO(Int p_set_svar_list, (void));
STATIC_PROTO(Int p_frozen_goals, (void));
STATIC_PROTO(Int p_all_frozen_goals, (void));
STATIC_PROTO(Int p_freeze_on_first, (void));
STATIC_PROTO(Int p_freeze, (void));
STATIC_PROTO(Int p_can_unify, (void));
STATIC_PROTO(Int p_non_ground, (void));

#ifdef COROUTINING

#include "corout_utils.h"

STATIC_PROTO(CELL Wake, (int, CELL, CELL));
STATIC_PROTO(sus_record *UpdateSVarList, (sus_record *));
STATIC_PROTO(sus_record *GetSVarList, (void));
STATIC_PROTO(CELL DerefSuspendedVar, (CELL));
STATIC_PROTO(void mark_sus_record, (sus_record *));
STATIC_PROTO(void mark_suspended_goal, (CELL *));
STATIC_PROTO(void AddSuspendedGoals, (sus_record *, sus_record *));
STATIC_PROTO(void ReleaseGoals, (sus_record *));
STATIC_PROTO(void wake_if_binding_vars_in_frozen_goal, (Term, sus_record *));
STATIC_PROTO(void AddSuspendedGoals, (sus_record *, sus_record *));
STATIC_PROTO(sus_record *has_been_suspended, (Term, sus_record *));
STATIC_PROTO(void AddSuspendedGoal, (Term, sus_record *));
STATIC_PROTO(Term AddSusToList, (Term, Term));
STATIC_PROTO(Term AddSusSubGoals, (Term, CELL *, int));
STATIC_PROTO(Int freeze_goal, (Term, Term));
STATIC_PROTO(Term AddVarIfNotThere, (Term, Term));
STATIC_PROTO(int can_unify_complex, (CELL *, CELL *, CELL *, Term *));
STATIC_PROTO(int can_unify, (Term, Term, Term *));
STATIC_PROTO(int non_ground_complex, (CELL *, CELL *, Term *));
STATIC_PROTO(int non_ground, (Term, Term *));
#ifdef FOLLOW_ENVIRONMENTS_FOR_SUSPENDED_GOALS
STATIC_PROTO(Term FindFrozenGoals, (Term, CELL *, int));
#endif

#ifdef MULTI_ASSIGNMENT_VARIABLES

inline static sus_record *
UpdateSVarList(sus_record *sl)
{
  /* make sl the new head of the suspension list, and update the list
     to use the old one. Note that the list is only bound once,
     MutableList is the one variable being updated all the time */
  return((sus_record *)UpdateTimedVar(MutableList, (CELL)sl));
}

inline static sus_record *
GetSVarList()
{
  Term t = ReadTimedVar(MutableList);
  /* just return the start of the list */
  if (t == TermNil)
    return(NULL);
  else
    return((sus_record *)t);
}

#endif

/* dif (and eventually others) may have the same goal suspended on the
   several variables. If this is the case, whenever we bind two
   variables we may need to wake the goals. That's implemented by
   going to the other guy's list, and checking if the same goal
   appears there.

*/

Term
ListOfWokenGoals(void) {
  sus_record *pt = (sus_record *)ReadTimedVar(WokenGoals);
  Term t;

  t = TermNil;
  while (pt->NR != (sus_record *)(&(pt->NR))) {
    t = MkPairTerm(pt->SG, t);
    pt = pt->NR;
  }
  t = MkPairTerm(pt->SG, t);  
  return(t);
}

static void ReleaseGoals(sus_record *from)
{
  /* follow the chain */
  sus_record *WGs = (sus_record *)ReadTimedVar(WokenGoals);

  if ((Term)WGs == TermNil) {
    UpdateTimedVar(WokenGoals, (CELL)from);
    /* from now on, we have to start waking up goals */
    CreepFlag = Unsigned(LCL0);
  } else {
    /* add to the end of the current list of suspended goals */
    CELL *where_to = (CELL *)SDeref((CELL)WGs);
    Bind(where_to, (CELL)from);
  }
}

static void
wake_if_binding_vars_in_frozen_goal(Term goal, sus_record *from)
{
  do {
    if (from->SG == goal) {
      sus_record *gf;

      /* A dif like goal has suspended on both variables. We cannot
	 wake it up directly, because it may have other goals
	 suspended on the same variable. So we'll just wake up a copy,
	 and wake up the copy.
	 */
      gf = (sus_record *)H;
      H += sizeof(sus_record)/sizeof(CELL);
      gf->NR = (sus_record *)&(gf->NR);
      gf->SG = goal;
#ifdef MULTI_ASSIGNMENT_VARIABLES
      gf->NS = UpdateSVarList(gf);
#endif
      ReleaseGoals(gf);
      /* done */
      return;
    }
    if (from->NR == (sus_record *)&(from->NR))
      return;
    else from = from->NR;
  } while (TRUE);
}

inline static void AddSuspendedGoals(sus_record *to, sus_record *from)
{
  /* deref the chain */
  do {
    if (IsApplTerm(to->SG))
      wake_if_binding_vars_in_frozen_goal(to->SG, from);
    if (to->NR == (sus_record *)&(to->NR))
      break;
    else to = to->NR;
  } while (TRUE);
  /* and bind it */
  Bind((CELL *)(to->NR), (CELL)from);
}


static sus_record *
has_been_suspended(Term goal, sus_record *from)
{
  do {
    if (from->SG == goal) {
      /* we found it */
      return (NULL);
    }
    if (from->NR == (sus_record *)&(from->NR))
      return (from);
    else from = from->NR;
  } while (TRUE);
  /* make lcc happy */
  return(NULL);
}

/* This is a simplified version for the case we add a goal to a
   suspended goal queue. It avoids having the same copy of the goal
   all over the place!
*/
inline static void AddSuspendedGoal(Term goal, sus_record *from)
{
  sus_record *gf;

  /* do nothing if we suspended before on the same goal! */
  if (IsApplTerm(goal) && ((from = has_been_suspended(goal, from)) == NULL))
    return;
  /* else add goal to the queue */
  gf = (sus_record *)H;
  H += sizeof(sus_record)/sizeof(CELL);
  gf->NR = (sus_record *)&(gf->NR);
  gf->SG = goal;
#ifdef MULTI_ASSIGNMENT_VARIABLES
  gf->NS = UpdateSVarList(gf);
#endif
  Bind((CELL *)&(from->NR), (CELL)gf);
}

static CELL
DerefSuspendedVar(CELL orig)
{
  register sus_tag *sreg = RepSuspendedVar(orig);
  Term t;

  sreg = deref_susp_chain(sreg);
  t = sreg->ActiveSus;
  if (!IsVarTerm(t) && IsAttachedTerm(t))
    return(attas[ExtFromFunctor(FunctorOfTerm(t))].deref_op(t));
  else
    return(t);
}

static void
mark_sus_record(sus_record *sg)
{
  if (MARKED(((CELL)(sg->NR))))
    return;
  MARK(((CELL *)&(sg->NR)));
  total_marked++;
  mark_variable((CELL *)&(sg->SG));
#ifdef MULTI_ASSIGNMENT_VARIABLES
  total_marked++;
  if (!IsAtomTerm((CELL)(sg->NS)))
    mark_suspended_goal((CELL *)(sg->NS));
  MARK(((CELL *)&(sg->NS)));
#endif
}

static void mark_suspended_goal(CELL *orig)
{
  register sus_tag *sreg = (sus_tag *)orig;

  if (MARKED(((CELL)(sreg->sus_fun))))
    return;
  MARK(((CELL *)&(sreg->sus_fun)));
  mark_sus_record(sreg->SG);
  mark_variable((CELL *)&(sreg->ActiveSus));
  total_marked += 2;
  MARK(((CELL *)&(sreg->SG)));
}

/*

   This routine does most of the work. It is called after
   someone tries to instantiate a suspension reference.

   Three operations are possible:

   SBIND:   trying to bind it to a constructed non-var term, most
	   often a primitive term;
   SISPAIR: the term is *going* to be bound to a list. We need to
	   return where.
   SISAPPL: the term is *going* to be bound to a compound term. We
	   need to return where, if we allow the binding.

*/

static CELL
Wake(int bind_op, CELL reg1, CELL reg2)
{
  
  /* if bound to someone else, follow until we find the last one */
  register sus_tag *susp = deref_susp_chain(RepSuspendedVar(reg1));
  /* next, do the operation */
  switch (bind_op) {

    /* Four cases:

       o Reg1 derefs to unbound, Reg2 derefs to unbound;
		Bind Two SVars
       o Reg1 derefs to unbound, Reg2 derefs to   bound;
		Bind SVar
       o Reg1 derefs to   bound, Reg2 derefs to unbound;
		Bind SVar
       o Reg1 derefs to   bound, Reg2 derefs to   bound;
		Unify two terms.

     */

  case SBIND:
    if (IsVarTerm(susp->ActiveSus)) {
      /* Time to bind a variable to a list of goals */
      /* reg2 should be bound to something */
    restart_sbind:
      if (IsAttachedTerm(reg2)) {
	Functor f;
	sus_tag *susp2;

	f = FunctorOfTerm(reg2);
	if (f != FunctorDelay) {
	  /* obtain the term */
	  reg2 = attas[ExtFromFunctor(f)].deref_op(reg2);
	  /* ignore reference chains */
	  goto restart_sbind;
	}

	/* Complicated case: we are binding two suspended vars
	   together!! */
	susp2 = deref_susp_chain(RepSuspendedVar(reg2));
	if (IsVarTerm(susp2->ActiveSus))
	  /* Case 1, we are really binding two susp variables together
	     ;-). */
	  {
	    /* Do the binding */
	    if (susp2->ActiveSus > susp->ActiveSus) {
	      /* susp is older */
	      /* bind */
	      Bind((CELL *)(susp2->ActiveSus), susp->ActiveSus);
	      /* and add to queue */
	      AddSuspendedGoals(susp->SG, susp2->SG);
	      /* give an OK! result */
	      return(TRUE);
	    } else {
	      /* susp2 is older */
	      /* bind */
	      Bind((CELL *)(susp->ActiveSus), susp2->ActiveSus);
	      /* and add to queue */
	      AddSuspendedGoals(susp2->SG, susp->SG);
	      /* give an OK! result */
	      return(TRUE);
	    }
	  }
	else  {
	  /* Case 2, reg2 had already been bound, so we just have to
	     wake up reg1 ;-) */
	  Bind((CELL *)(susp->ActiveSus), susp2->ActiveSus);
	  /* now, we can wake up a few goals */
	  ReleaseGoals(susp->SG);
	  /* say we are ok! */
	  return(TRUE);
	}
      } else {
	  /* case 2, reg2 had already been bound, so we just have to
	     wake up reg1 */
	  Bind((CELL *)(susp->ActiveSus), reg2);
	  /* now, we can wake up a few goals */
	  ReleaseGoals(susp->SG);
	  /* say we are ok! */
	  return(TRUE);
      }
    } else {
      CELL reg1 = susp->ActiveSus;
    restart_sbind_of_reg2:
      /* reg1 had been bound before! Just look at reg2 in case it is
	 itself a sref */
      if (IsAttachedTerm(reg2)) {
	Functor f;
	sus_tag *susp2;

	f = FunctorOfTerm(reg2);
	if (f != FunctorDelay) {
	  /* obtain the term */
	  reg2 = attas[ExtFromFunctor(f)].deref_op(reg2);
	  /* ignore reference chains */
	  goto restart_sbind_of_reg2;
	}
	susp2 = deref_susp_chain(RepSuspendedVar(reg2));
	/* Case 3, reg1 had already been bound, so we just have to
	   wake up reg2 ;-) */
	if (IsVarTerm(susp2->ActiveSus))
	  {
	    /* release the variable into the WokenGoals list */
	    ReleaseGoals(susp2->SG);
	    /* bind it to t1's value */
	    Bind((CELL *)(susp2->ActiveSus), reg1);
	    /* say we are ok! */
	    return(TRUE);
	  }
	/* continue with reg2 */
	else reg2 = susp2->ActiveSus;
      }
      /* Case 4, we got rid of SVars.  */
      /* reg2 has the value of the second variable */
      if (reg1 != reg2)
	return(IUnify(reg1, reg2));
      else
	return(TRUE);
    }
    break;

  case SISPAIR:
    /* we found a variable which is going to be bound to something */
    if (IsVarTerm(susp->ActiveSus)) {
      /* release the variable into the WokenGoals list */
      ReleaseGoals(susp->SG);
      /* and return the variable as the result of dereferencing */
      /* it will become a pair */
      return(susp->ActiveSus);
    } else if (IsPairTerm(susp->ActiveSus))
      return(susp->ActiveSus);
      /* oopss: it had become something very different, should fail now! */
    else return(FALSE);
    break;

  case SISAPPL:
    /* we found a variable which is going to be bound to something */
    if (IsVarTerm(susp->ActiveSus)) {
      /* release the variable into the WokenGoals list */
      ReleaseGoals(susp->SG);
      /* and return the variable as the result of dereferencing */
      /* it will become a struct */
      return(susp->ActiveSus);
    } else if (IsApplTerm(susp->ActiveSus))
      return(susp->ActiveSus);
      /* oopss: it had become something very different, should fail now! */
    else return(FALSE);
    break;
  }
  return(0);
}

/* find all goals frozen in the current chain of environments */

/* This will also mark them as bound, in order that goal lists
   won't be displayed twice */
static Term
AddSusToList(Term t, Term t1)
{
  sus_tag * susp;
  if (IsVarTerm((t1 = (susp =
		deref_susp_chain(RepSuspendedVar(t1)))->ActiveSus))) {
    /* we found an active suspension variable */
    sus_record *s = susp->SG;
    while (s->NR != (sus_record *)&(s->NR)) {
      t = MkPairTerm(s->SG,t);
      s = s->NR;
    } while (s->NR != (sus_record *)&(s->NR));
    t = MkPairTerm(s->SG,t);
    Bind((CELL *)(susp->ActiveSus), TermNil);
  } else if (IsApplTerm(t1)) {
    int args = ArityOfFunctor(FunctorOfTerm(t1));
    t = AddSusSubGoals(t, RepAppl(t1)+1, args);
  } else if (IsPairTerm(t1)) {
    t = AddSusSubGoals(t, RepPair(t1), 2);
  }
  return(t);
}

/* used to search from subarguments from within a compound term */
static Term
AddSusSubGoals(Term t, CELL *saved_var, int max)
{
  int i;

  for (i = 0; i < max; i++)
    {
      Term t1 = SDerefa(saved_var);
      if (!IsVarTerm(t1)) {
	if (IsApplTerm(t1)) {
	  Functor f = FunctorOfTerm(t1);
	  
	  if (f == FunctorDelay) {
	    t = AddSusToList(t, t1);
	  } else {
	    int args = ArityOfFunctor(f);

	    t = AddSusSubGoals(t, RepAppl(t1)+1, args);
	  }
	} else if (IsPairTerm(t1)) {
	  t = AddSusSubGoals(t, RepPair(t1), 2);
	}
      }
    }
  return(t);
}

static Int
freeze_goal(Term t, Term g)
{
  if (IsVarTerm(t)) {
    sus_record *gf;
    sus_tag *vs;

    /* create a new suspension record */
    gf = (sus_record *)H;
    /* I assume here sus_record has size multiple of CELL !!!! */
    H += sizeof(sus_record)/sizeof(CELL);
    gf->NR = (sus_record *)&(gf->NR);
    gf->SG = g;
#ifdef MULTI_ASSIGNMENT_VARIABLES
    gf->NS = UpdateSVarList(gf);
#endif
    vs = (sus_tag *)(H);
    H += sizeof(sus_tag)/sizeof(CELL);
    vs->sus_fun = FunctorDelay;
    vs->SG = gf;
    vs->ActiveSus = (CELL)&(vs->ActiveSus);
    Bind((CELL *)t,AbsSuspendedVar(vs))
    return(TRUE);
  }
  else {
    if (IsAttachedTerm(t)) {
      Functor f;
      sus_tag *susp;

      f = FunctorOfTerm(t);
      if (f != FunctorDelay) {
	/* obtain the term */
	t = attas[ExtFromFunctor(f)].deref_op(t);
	/* ignore reference chains */
	return(freeze_goal(t, g));
      }

      susp = deref_susp_chain(RepSuspendedVar(t));
      if (IsVarTerm(susp->ActiveSus)) {
	AddSuspendedGoal(g, susp->SG);
	return(TRUE);
      }
    }
    /* Oops, first argument was bound :-( */
    Error(TYPE_ERROR_VARIABLE, t, "freeze/2");
    return(FALSE);
  }
}

#endif /* COROUTINING */

static Int
p_read_svar_list(void)
{
#ifdef COROUTINING
#ifdef MULTI_ASSIGNMENT_VARIABLES
  if (!unify(ARG1, MutableList))
    return(FALSE);
#endif
#endif
  return(TRUE);
}

static Int
p_set_svar_list(void)
{
#ifdef COROUTINING
#ifdef MULTI_ASSIGNMENT_VARIABLES
  MutableList = Deref(ARG1);
#endif
#endif
  return(TRUE);
}

static Int
p_freeze(void)
{
#ifdef COROUTINING
  Term t = SDeref(ARG1);
  return(freeze_goal(t, Deref(ARG2)));
#else
  return(FALSE);
#endif /* COROUTINING */
}

/* The idea here is that we are trying to freeze on a list of
   variables. If we can freeze on the first one, we create a
   suspension record and are off to see the wizard of Oz. Otherwise,
   the goal fails, indicating we did not have to freeze (look at code
   for wait and for block to understand why.
*/
static Int p_freeze_on_first(void)
{
#ifdef COROUTINING
  Term r = SDeref(ARG1);
  int i;
  CELL *pt;

  if (!IsApplTerm(r)) return(FALSE);
  i = ArityOfFunctor(FunctorOfTerm(r));
  pt = RepAppl(r)+1;
  do {
    if (IsNonVarTerm(Derefa(pt)))
      return(FALSE);
    i --;
    pt++;
  } while(i);
  /* we can freeze on the first variable */
  return(freeze_goal(SDerefa(RepAppl(r)+1), Deref(ARG2)));
#else
  return(FALSE);
#endif
}


/* return a queue with goals currently frozen on the first argument */
static Int p_frozen_goals(void)
{
  /* initially, we do not know of any goals frozen */
  Term t = TermNil;
#ifdef COROUTINING
  Term t1 = SDeref(ARG1);
  CELL *pt1, *pt0;
  /* make B and HB point to H to guarantee all bindings will
     be trailed
     */
  pt1 = (CELL *)B;
  pt0 = TR;
  HB = H;
  B = (choiceptr)H;
  /* look at the first argument */
  if (!IsVarTerm(t1)) {
    if (IsApplTerm(t1)) {
      Functor f = FunctorOfTerm(t1);
	  
      if (f == FunctorDelay) {
	t = AddSusToList(t, t1);
      } else {
	int args = ArityOfFunctor(f);
	t = AddSusSubGoals(t, RepAppl(t1)+1, args);
      }
    } else if (IsPairTerm(t1)) {
      t = AddSusSubGoals(t, RepPair(t1), 2);
    }
  }
  B = (choiceptr)pt1;
  /* untrail all bindings made by IUnify */
  while (TR != pt0) {
    pt1 = (CELL *)(*--TR);
    RESET_VARIABLE(pt1);
  }
  HB = B->cp_h;
#endif
  return(unify(ARG2,t));
}

/* return a queue with all goals frozen in the system */
static Int p_all_frozen_goals(void)
{
  /* initially, we do not know of any goals frozen */
  Term t = TermNil;
#ifdef COROUTINING
#ifdef MULTI_ASSIGNMENT_VARIABLES
  sus_record *x = GetSVarList();
  if (x == NULL)
    return(unify(ARG1,t));
  /* okay, we are on top of the list of variables. Let's burn rubber!
   */
  while ((CELL)x != TermNil) {
    t = MkPairTerm(x->SG,t);
    x = x->NS;
  }
#else
  return(unify(ARG1,t));
#endif
#endif
  return(unify(ARG1,t));
}

#ifdef COROUTINING

#ifdef RATIONAL_TREES

#define UNWIND_CUNIF()                                        \
         while (visited < AuxSp) {                            \
            CELL *pt1 = (CELL *)visited[0];                   \
            *pt1 = visited[1];                                \
            visited += 2;                                     \
         }

#else
#define UNWIND_CUNIF()
#endif

/* check if variable was there */
static Term AddVarIfNotThere(Term var , Term dest)
{
  Term test = dest;
  while (test != TermNil) {
    if ((RepPair(test))[0] == var) return(dest);
    else test = (RepPair(test))[1];
  }
  return(MkPairTerm(var,dest));
}


/* This routine verifies whether two complex structures can unify. */
static int can_unify_complex(register CELL *pt0,
		register CELL *pt0_end,
		register CELL *pt1,
		Term  *Vars)
{

  /* This is really just unification, folks */
  CELL *saved_TR, *saved_HB;
  choiceptr saved_B;

#ifdef RATIONAL_TREES
  register CELL *visited = AuxSp;
#endif
  register CELL **to_visit = (CELL **)PreAllocCodeSpace();
  CELL **to_visit_base = to_visit;

  /* make sure to trail all bindings */
  saved_TR = TR;
  saved_B = B;
  saved_HB = HB;
  HB = H;

 loop:
  while (pt0 < pt0_end) {
    register CELL d0, d1;
    ++ pt0;
    ++ pt1;
    d0 = SDerefa(pt0);
    d1 = SDerefa(pt1);
  restart:
    if (IsVarTerm(d0)) {
      if (IsVarTerm(d1)) {
	if (d0 != d1) {
	  /* we need to suspend on both variables ! */
	  *Vars = AddVarIfNotThere(d0, AddVarIfNotThere(d1,*Vars));
	  /* bind the two variables, we would have to do that to unify
	     them */
	  if (d1 > d0) {
	    Bind((CELL *)d1, d0);
	  } else {
	    Bind((CELL *)d0, d1);
	  }
	}
	/* continue the loop */
	continue;
      }
      else {
	if (IsAttachedTerm(d1)) {

	  Functor f = FunctorOfTerm(d1);

	  if (f != FunctorDelay) {
	    d1 = (attas[ExtFromFunctor(f)].deref_op)(d1);
	    goto restart;
	  } else {
	    sus_tag *susp1;

	    /* d0 is a true variable, only d1 is a suspended
	       variable. Hence, they can't be the same variable. */
	    susp1 = deref_susp_chain(RepSuspendedVar(d1));
	    if (IsVarTerm(susp1->ActiveSus)) {
	      *Vars = AddVarIfNotThere(d0,AddVarIfNotThere(d1,*Vars));
	    } else {
	      /* false alarm */
	      *Vars = AddVarIfNotThere(d0,*Vars);
	    }
	  }
	} else {
	  /* oh no, some more variables! */
	  *Vars = AddVarIfNotThere(d0, *Vars);
	}
      }
      /* now bind it */
      Bind((CELL *)d0, d1);
      /* continue the loop */
    } else if (IsVarTerm(d1))  {
      if (IsAttachedTerm(d0)) {
	/* d1 is a true variable, only d0 is a suspended
	   variable. Hence, they can't be the same variable. */

	Functor f = FunctorOfTerm(d0);

	if (f != FunctorDelay) {
	  d0 = (attas[ExtFromFunctor(f)].deref_op)(d0);
	  goto restart;
	} else {
	  sus_tag *susp0;

	  susp0 = deref_susp_chain(RepSuspendedVar(d0));
	  if (IsVarTerm(susp0->ActiveSus)) {
	    *Vars = AddVarIfNotThere(d0,AddVarIfNotThere(d1,*Vars));
	  } else {
	    /* false alarm */
	    *Vars = AddVarIfNotThere(d1,*Vars);
	  }
	}
      } else {
	*Vars = AddVarIfNotThere(d1, *Vars);
      }
      /* and bind it */
      Bind((CELL *)d1, d0);
      /* continue the loop */
    } else {
      if (d0 == d1) continue;
      if (IsAttachedTerm(d0)) {
	Functor f = FunctorOfTerm(d0);

	if (f != FunctorDelay) {
	  d0 = (attas[ExtFromFunctor(f)].deref_op)(d0);
	  goto restart;
	} else {
	  sus_tag *susp0;

	  susp0 = deref_susp_chain(RepSuspendedVar(d0));
	  if (IsVarTerm(susp0->ActiveSus)) {
	    if (IsAttachedTerm(d1)) {
	      sus_tag *susp1;

	      if (FunctorOfTerm(d0) != FunctorDelay) {
		/* obtain the term */
		d0 = attas[ExtFromFunctor(FunctorOfTerm(d0))].deref_op(d0);
		/* ignore reference chains */
		goto restart;
	      }

	      /* find out what's going on with this variable */
	      susp1 = deref_susp_chain(RepSuspendedVar(d1));
	      if (IsVarTerm(susp1->ActiveSus)) {
		if (susp1->ActiveSus != susp0->ActiveSus) {
		  /* we have two suspended variables, so we unify them */
		  *Vars = AddVarIfNotThere(d0,AddVarIfNotThere(d1,*Vars));
		  if (susp1->ActiveSus > susp0->ActiveSus) {
		    Bind((CELL *)(susp1->ActiveSus), susp0->ActiveSus);
		  } else {
		    Bind((CELL *)(susp0->ActiveSus), susp1->ActiveSus);
		  }
		}
		continue;
	      } else {
		/* the suspended variable binds to the term */
		*Vars = AddVarIfNotThere(d0, *Vars);
		Bind((CELL *)(susp0->ActiveSus), susp1->ActiveSus);
		continue;
	      }
	    } else {
	      /* d0 is sus, d1 is bound, bind d0 to d1 */
	      *Vars = AddVarIfNotThere(d0, *Vars);
	      Bind((CELL *)(susp0->ActiveSus), d1);
	      continue;
	    }
	  } else {
	    /* otherwise d0 is actually bound, and we may feel free to
	       continue */
	    d0 = susp0->ActiveSus;
	  }
	}
      }
      if (IsAttachedTerm(d1)) {

	Functor f = FunctorOfTerm(d1);

	if (f != FunctorDelay) {
	  d1 = (attas[ExtFromFunctor(f)].deref_op)(d1);
	  goto restart;
	} else {
	  /* we know d0 is bound */
	  sus_tag *susp1;

	  susp1 = deref_susp_chain(RepSuspendedVar(d1));
	  if (IsVarTerm(susp1->ActiveSus)) {
	    /* d1 is a SVar! */
	    /* just do as if you were going to bind it */
	    *Vars = AddVarIfNotThere(d1, *Vars);
	    Bind((CELL *)(susp1->ActiveSus), d0);
	    continue;
	  } else {
	    d1 = susp1->ActiveSus;
	  }
	}
      }
      if (IsAtomOrIntTerm(d0) || IsAtomOrIntTerm(d1)) {
	  if (d0 != d1) goto comparison_failed;
	/* else continue the loop */
      }
      else if (IsPairTerm(d0)) {
	if (!IsPairTerm(d1)) goto comparison_failed;
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	if (d0 > d1) {
	  visited -= 2;
	  visited[0] = (CELL)pt0;
	  visited[1] = *pt0;
	  *pt0 = d1;
	} else {
	  visited -= 2;
	  visited[0] = (CELL)pt1;
	  visited[1] = *pt1;
	  *pt1 = d0;
	}
#endif
	/* store the terms to visit */
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit[2] = pt1;
	  to_visit += 3;
	}
	pt0 = RepPair(d0) - 1;
	pt0_end = RepPair(d0) + 1;
	pt1 = RepPair(d1) - 1;
	continue;
      }
      else if (IsApplTerm(d0)) {
	register Functor f;
	register CELL *ap2, *ap3;
	if (!IsApplTerm(d1)) {
	  goto comparison_failed;
	} else {
	  /* store the terms to visit */
	  ap2 = RepAppl(d0);
	  ap3 = RepAppl(d1);
	  f = (Functor)(*ap2);
	  /* compare functors */
	  if (f != (Functor)*ap3) {
	    goto comparison_failed;
	  }
	  if (IsExtensionFunctor(f)) {
	    if (f == FunctorDouble) {
	      if (FunctorOfTerm(d1) == FunctorDouble &&
		  FloatOfTerm(d0) == FloatOfTerm(d1))
		continue;
	      else
		goto comparison_failed;
	    } else if (f == FunctorGeneralRef) {
	      if (FunctorOfTerm(d1) == FunctorGeneralRef &&
		  FloatOfTerm(d0) == FloatOfTerm(d1))
		continue;
	      else
		goto comparison_failed;
	    } else if (f == FunctorDBRef) {
	      if (d0 == d1)
		continue;
	      else
		goto comparison_failed;
	    } else if (f == FunctorLongInt) {
	      if (FunctorOfTerm(d1) == FunctorLongInt &&
		  LongIntOfTerm(d0) == LongIntOfTerm(d1))
		continue;
	      else
		goto comparison_failed;
	    }
	  }
#ifdef RATIONAL_TREES
	  /* now link the two structures so that no one else will */
	  /* come here */
	  if (d0 > d1) {
	    visited -= 2;
	    visited[0] = (CELL)pt0;
	    visited[1] = *pt0;
	    *pt0 = d1;
	  } else {
	    visited -= 2;
	    visited[0] = (CELL)pt1;
	    visited[1] = *pt1;
	    *pt1 = d0;
	  }
#endif
	  /* store the terms to visit */
	  if (pt0 < pt0_end) {
	    to_visit[0] = pt0;
	    to_visit[1] = pt0_end;
	    to_visit[2] = pt1;
	    to_visit += 3;
	  }
	  d0 = ArityOfFunctor(f);
	  pt0 = ap2;
	  pt0_end = ap2 + d0;
	  pt1 = ap3;
	  continue;
	}
      }

    }

  }
  /* Do we still have compound terms to visit */
  if (to_visit > (CELL **)to_visit_base) {
    to_visit -= 3;
    pt0 = to_visit[0];
    pt0_end = to_visit[1];
    pt1 = to_visit[2];
    goto loop;
  }
  /* success */
  UNWIND_CUNIF();
  /* restore B, and later HB */
  B = saved_B;
  HB = saved_HB;
  /* untrail all bindings made by IUnify */
  while (TR != saved_TR) {
    pt1 = (CELL *)(*--TR);
    RESET_VARIABLE(pt1);
  }
  return(TRUE);

 comparison_failed:
  /* failure */
  UNWIND_CUNIF();
  /* restore B, and later HB */
  B  = saved_B;
  HB = saved_HB;
  /* the system will take care of TR for me, no need to worry here! */
  return(FALSE);
}

static int
can_unify(Term t1, Term t2, Term *Vars)
{
  t1 = SDeref(t1);
  t2 = SDeref(t2);
restart:
  if (t1 == t2)
    return (TRUE);
  if (IsVarTerm(t1)) {
    /* we know for sure  they can't be different */
    if (IsVarTerm(t2)) {
      /* we need to suspend on both variables ! */
      *Vars = MkPairTerm(t1,MkPairTerm(t2,TermNil));
      return(TRUE);
    }
    else {
      if (IsAttachedTerm(t2)) {
	Functor f = FunctorOfTerm(t2);

	if (f != FunctorDelay) {
	  t2 = (attas[ExtFromFunctor(f)].deref_op)(t2);
	  goto restart;
	} else {
	  /* t1 is a true variable, only t2 is a suspended
	     variable. Hence, they can't be the same variable. */
	  sus_tag *susp2;

	  susp2 = deref_susp_chain(RepSuspendedVar(t2));
	  if (IsVarTerm(susp2->ActiveSus)) {
	    *Vars = MkPairTerm(t1,MkPairTerm(t2,TermNil));
	  } else {
	    /* false alarm */
	    *Vars = MkPairTerm(t1,TermNil);
	  }
	}
      } else {
	/* wait until t1 is bound */
	*Vars = MkPairTerm(t1,TermNil);
      }
      return(TRUE);
    }
  } else if (IsVarTerm(t2)) {
    if (IsAttachedTerm(t1)) {
      Functor f = FunctorOfTerm(t1);

      if (f != FunctorDelay) {
	  t1 = (attas[ExtFromFunctor(f)].deref_op)(t1);
	  goto restart;
      } else {
	/* t2 is a true variable, only t1 is a suspended
	   variable. Hence, they can't be the same variable. */
	sus_tag *susp1;

	susp1 = deref_susp_chain(RepSuspendedVar(t1));
	if (IsVarTerm(susp1->ActiveSus)) {
	  *Vars = MkPairTerm(t1,MkPairTerm(t2,TermNil));
	} else {
	  /* false alarm */
	  *Vars = MkPairTerm(t2,TermNil);
	}
      }
      return(TRUE);
    } else {
      /* wait until t2 is bound */
      *Vars = MkPairTerm(t2,TermNil);
      return(TRUE);
    }
  }
  if (IsAttachedTerm(t1)) {
    Functor f = FunctorOfTerm(t1);
    sus_tag *susp1;

    if (f != FunctorDelay) {
      t1 = (attas[ExtFromFunctor(f)].deref_op)(t1);
      goto restart;
    }

    susp1 = deref_susp_chain(RepSuspendedVar(t1));
    if (IsVarTerm(susp1->ActiveSus)) {
      if (IsAttachedTerm(t2)) {
	Functor f = FunctorOfTerm(t1);
	sus_tag *susp2;
	
	if (f != FunctorDelay) {
	  t1 = (attas[ExtFromFunctor(f)].deref_op)(t1);
	  goto restart;
	}
	/* find out what's going on with this variable */

	susp2 = deref_susp_chain(RepSuspendedVar(t2));
	if (IsVarTerm(susp2->ActiveSus)) {
	  if (susp2->ActiveSus != susp1->ActiveSus)
	    /* we have two suspended variables, so we could unify */
	    *Vars = MkPairTerm(t1,MkPairTerm(t2,TermNil));
	  return(TRUE);
	} else {
	  /* the suspended variable would bind to the term */
	  *Vars = MkPairTerm(t1, TermNil);
	  return(TRUE);
	}
      } else {
	/* t1 is suspended, t2 is bound, we would bind t1 to t2 */
	*Vars = MkPairTerm(t1, TermNil);
	return(TRUE);
      }
    } else {
      /* otherwise t1 is actually bound, and we may feel free to
	 continue */
      t1 = susp1->ActiveSus;
    }
  }
  if (IsAttachedTerm(t2)) { 
    Functor f = FunctorOfTerm(t2);

    if (f != FunctorDelay) {
      t2 = (attas[ExtFromFunctor(f)].deref_op)(t2);
      goto restart;
    } else {
      /* we know for sure t2 is a suspended variable, and that t1
	 is bound */
      sus_tag *susp2;

      susp2 = deref_susp_chain(RepSuspendedVar(t2));
      if (IsVarTerm(susp2->ActiveSus)) {
	/* t2 is really a SVar */
	/* just do as if you were going to bind it */
	*Vars = MkPairTerm(t2, TermNil);
	return(TRUE);
      } else {
	t2 = susp2->ActiveSus;
      }
    }
  }
  /* Two standard terms at last! */
  if (IsAtomOrIntTerm(t1) || IsAtomOrIntTerm(t2)) {
    /* Two primitive terms can only be equal if they are
       the same. If they are, $eq succeeds without further ado.
       */
    if (t1 != t2)
      return(FALSE);
    else {
      *Vars = TermNil;
      return(TRUE);
    }
  } else if (IsPairTerm(t1)) {
    if (IsPairTerm(t2)) {
      return(can_unify_complex(RepPair(t1)-1, RepPair(t1)+1,
			       RepPair(t2)-1, Vars));
    } else return(FALSE);
  } else {
    Functor f = FunctorOfTerm(t1);
    if (f != FunctorOfTerm(t2))
      return (FALSE);
    /* Two complex terms with the same functor */
    if (!IsExtensionFunctor(f))
      return(can_unify_complex(RepAppl(t1),
			       RepAppl(t1)+ArityOfFunctor(f),
			       RepAppl(t2), Vars));
    else
      if (f == FunctorDouble) {
	if (FunctorOfTerm(t2) == FunctorDouble &&
	    FloatOfTerm(t1) == FloatOfTerm(t2))
	  return(TRUE);
	else
	  return(FALSE);
      } else if (f == FunctorDBRef) {
	if (t1 == t2)
	  return(TRUE);
	else
	  return(FALSE);
      } else if (f == FunctorGeneralRef) {
	if (FunctorOfTerm(t2) == FunctorGeneralRef &&
	    RepAppl(t1)[1] == RepAppl(t2)[1])
	  return(TRUE);
	else
	  return(FALSE);
      } else if (f == FunctorLongInt) {
	if (FunctorOfTerm(t2) == FunctorLongInt &&
	    LongIntOfTerm(t1) == LongIntOfTerm(t2))
	  return(TRUE);
	else
	  return(FALSE);
      } else return(FALSE);
  }
}

/* This routine verifies whether a complex has variables. */
static int non_ground_complex(register CELL *pt0,
		register CELL *pt0_end,
		Term  *Var)
{

#ifdef RATIONAL_TREES
  register CELL *visited = AuxSp;
#endif
  register CELL **to_visit = (CELL **)PreAllocCodeSpace();
  CELL **to_visit_base = to_visit;

 loop:
  while (pt0 < pt0_end) {
    register CELL d0;
    ++ pt0;
    d0 = SDerefa(pt0);
    if (IsVarTerm(d0)) {
      *Var = d0;
      goto var_found;
    }
  restart_loop:    
    if (IsPairTerm(d0)) {
#ifdef RATIONAL_TREES
      /* now link the two structures so that no one else will */
      /* come here */
      visited -= 2;
      visited[0] = (CELL)pt0;
      visited[1] = *pt0;
      *pt0 = TermNil;
#endif
      /* store the terms to visit */
      if (pt0 < pt0_end) {
	to_visit[0] = pt0;
	to_visit[1] = pt0_end;
	to_visit += 2;
      }
      pt0 = RepPair(d0) - 1;
      pt0_end = RepPair(d0) + 1;
    }
    else if (IsApplTerm(d0)) {
      register Functor f;
      register CELL *ap2;

      /* store the terms to visit */
      ap2 = RepAppl(d0);
      f = (Functor)(*ap2);

      if (IsAttachFunc(f)) {
	Term d1 = attas[ExtFromFunctor(f)].deref_op(d0);
	if (IsVarTerm(d0)) {
	  *Var = d0;
	  goto var_found;
	} else
	  d0 = d1;
	goto restart_loop;
      }
#ifdef RATIONAL_TREES
      /* now link the two structures so that no one else will */
      /* come here */
      visited -= 2;
      visited[0] = (CELL)pt0;
      visited[1] = *pt0;
      *pt0 = TermNil;
#endif
      /* store the terms to visit */
      if (pt0 < pt0_end) {
	to_visit[0] = pt0;
	to_visit[1] = pt0_end;
	to_visit += 2;
      }
      d0 = ArityOfFunctor(f);
      pt0 = ap2;
      pt0_end = ap2 + d0;
    }
    /* just continue the loop */
  }

  /* Do we still have compound terms to visit */
  if (to_visit > (CELL **)to_visit_base) {
    to_visit -= 2;
    pt0 = to_visit[0];
    pt0_end = to_visit[1];
    goto loop;
  }

  /* the term is ground */
  UNWIND_CUNIF();
  return(FALSE);

 var_found:
  /* the term is non-ground */
  UNWIND_CUNIF();
  /* the system will take care of TR for me, no need to worry here! */
  return(TRUE);
}

static int
non_ground(Term t, Term *Var)
{
  t = SDeref(t);
  if (IsVarTerm(t)) {
    /* we found a variable */
    *Var = t;
    return(TRUE);
  }
restart_loop:
  if (IsPrimitiveTerm(t)) {
    return(FALSE);
  } else if (IsPairTerm(t)) {
    return(non_ground_complex(RepPair(t)-1, RepPair(t)+1, Var));
  } else {
    Functor f = FunctorOfTerm(t);
    if (IsAttachFunc(f)) {
      Term ti = attas[ExtFromFunctor(f)].deref_op(t);
      if (IsVarTerm(ti)) {
	*Var = t;
	return(TRUE);
      }
      t = ti;
      goto restart_loop;
    }
    return(non_ground_complex(RepAppl(t),
			  RepAppl(t)+ArityOfFunctor(FunctorOfTerm(t)),
			      Var));
  }
}

#endif

/* check whether the two terms unify and return what variables should
   be bound before the terms are execatly equal */
static Int p_can_unify(void)
{
#ifdef COROUTINING
  Term r = TermNil;
  if (!can_unify(ARG1, ARG2, &r))
    return(FALSE);
  return (unify(ARG3, r));
#else
  return(FALSE);
#endif
}

/* if the term is not ground return a variable in the term */
static Int p_non_ground(void)
{
#ifdef COROUTINING
  Term r;
  if (!non_ground(ARG1, &r))
    return(FALSE);
  return (unify(ARG2, r));
#else
  return(FALSE);
#endif
}

void InitCoroutPreds(void)
{
#ifdef COROUTINING
  Atom            at;
  PredEntry      *pred;

  attas[susp_ext].bind_op = Wake;
  attas[susp_ext].deref_op = DerefSuspendedVar;
  attas[susp_ext].binding_transparent = TRUE;
  attas[susp_ext].mark_op = mark_suspended_goal;
  at = LookupAtom("$wake_up_goal");
  pred = RepPredProp(PredProp(at, 2));
  WakeUpCode = (CELL *) & (pred->CodeOfPred);
#endif /* COROUTINING */
  InitCPred("$read_svar_list", 1, p_read_svar_list, SafePredFlag);
  InitCPred("$set_svar_list", 1, p_set_svar_list, SafePredFlag);
  InitCPred("$freeze", 2, p_freeze, 0);
  InitCPred("freeze_on_first", 2, p_freeze_on_first, SafePredFlag|TestPredFlag);
  InitCPred("$frozen_goals", 2, p_frozen_goals, SafePredFlag);
  InitCPred("$all_frozen_goals", 1, p_all_frozen_goals, SafePredFlag);
  InitCPred("$can_unify", 3, p_can_unify, SafePredFlag);
  InitCPred("$non_ground", 2, p_non_ground, SafePredFlag);
}


