/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		dbase.c							 *
* Last rev:	8/2/88							 *
* mods:									 *
* comments:	YAP's internal data base				 *
*									 *
*************************************************************************/
#ifdef SCCS
static char     SccsId[] = "%W% %G%";
#endif

#include "Yap.h"
#include "clause.h"
#include "heapgc.h"
#include "io.h"

/* There are two options to implement traditional immediate update semantics.

   - In the first option, we only remove an element of the chain when
   it is phisically disposed of. This simplifies things, because
   pointers are always valid, but it complicates some stuff a bit:

   o You may have go through long lines of deleted db entries before you 
   actually reach the one you want.

   o Deleted clauses are also not removed of the chain. The solution
   was to place a fail in every clause, but you still have to
   backtrack through failed clauses.

   An alternative solution is to remove clauses from the chain, even
   if they are still phisically present. Unfortunately this creates
   problems because immediate update semantics means you have to
   backtrack clauses or see the db entries stored later. 

   There are several solutions. One of the simplest is to use an age
   counter. When you backtrack to a removed clause or to a deleted db
   entry you use the age to find newly entered clauses in the DB.

   This still causes a problem when you backtrack to a deleted
   clause, because clauses are supposed to point to the next
   alternative, and having been removed from the chain you cannot
   point there directly. One solution is to have a predicate in C that 
   recovers the place where to go to and then gets rid of the clause.

*/



#ifdef  KEEP_ENTRY_AGE
#define DISCONNECT_OLD_ENTRIES 1
#else
#define KEEP_OLD_ENTRIES_HANGING_ABOUT 1
#endif  /* KEEP_ENTRY_AGE */

#ifdef MACYAPBUG
#define Register
#else
#define Register	register
#endif

/* Flags for recorda or recordz				 */
#define MkFirst	1
#define MkLast	2
#define MkCode  4
#define WithRef	8
#define MkIfNot	16
#define InQueue	32

#define FindDBProp(V,Y)	( GetDBProp(V,Y) )
#define FrstDBRef(V)	( (V) -> First )
#define NextDBRef(V)	( (V) -> Next )

#define DBLength(V)	(sizeof(DBStruct) + (Int)(V) + CellSize)
#define AllocDBSpace(V)	((DBRef)AllocCodeSpace(V))
#define FreeDBSpace(V)	FreeCodeSpace(V)

#define NO_ERROR_IN_DB 0
#define OVF_ERROR_IN_DB 1
#define SOVF_ERROR_IN_DB 2
#define OTHER_ERROR_IN_DB 3

#if SIZEOF_INT_P==4
#define ToSmall(V)	((BITS16)(Unsigned(V)>>2))
#else
#define ToSmall(V)	((BITS16)(Unsigned(V)>>3))
#endif

#ifdef SFUNC

#define MaxSFs		256

typedef struct {
  Term            SName;	/* The culprit */
  CELL           *SFather;      /* and his father's position */
}               SFKeep;
#endif

typedef struct
{
  Functor id;		/* identify this as being pointed to by a DBRef */
  SMALLUNSGN    Flags;  /* always required */
  DBRef FirstInQueue, LastInQueue;
}  db_queue;

/*
 * These masks are used before the unify, for a quick check of the structure 
 */
#define MListVar	0x00
#define MAtom		0x80
#define MStruc		0xc0
#define MNum		0x40

#define HashMask	0x3f
#define FullHashMask	0xff

#define SHIFT_FACTOR    (sizeof(CELL))

#define AtomHash(t)	((Unsigned(t)>>4)&HashMask)
#define FunctorHash(t)  ((Unsigned(t)>>4)&HashMask)
#define FullFunctorHash(t)  ((Unsigned(t)>>4)&FullHashMask)
#define NumberHash(t)    (Unsigned(IntOfTerm(t))&HashMask)
#define GlobNumHash(t)  ((RepAppl(t)[1]) & HashMask)
#define RefHash(t)	(((CELL)DBRefOfTerm(t)) & HashMask)

/* traditionally, YAP used a link table to recover IDB terms*/
#define IDB_LINK_TABLE 1
/* a second alternative is to just use a tag */
/*#define IDB_USE_MBIT 1*/

/* These global variables are necessary to build the data base
   structure */
#ifdef IDB_LINK_TABLE
static BITS16  *lr = NULL, *LinkAr;
#endif
static int      DBErrorFlag = FALSE; /* error while recording */
/* we cannot call Error directly from within recorded(). These flags are used
   to delay for a while
*/
static yap_error_number DBErrorNumber; /* error number */
static yap_error_number DBErrorTerm; /* error term */
static char     *DBErrorMsg;       /* Error Message */
static DBRef    *tofref;	/* place the refs also up	 */

static SMALLUNSGN      DBModule;

CELL *next_float = NULL;

#ifdef SFUNC
static CELL    *FathersPlace;	/* Where the father was going when the term
				 * was reached */
static SFKeep  *SFAr, *TopSF;	/* Where are we putting our SFunctors */
#endif

static DBRef    found_one;	/* Place where we started recording */

#ifdef SUPPORT_HASH_TABLES
typedef struct {
  CELL  key;
  DBRef entry;
} hash_db_entry;

typedef table {
  Int NOfEntries;
  Int HashArg;
  hash_db_entry *table;
} hash_db_table;
#endif

STATIC_PROTO(CELL *cpcells,(CELL *,CELL*,Int));
#ifdef IDB_LINK_TABLE
STATIC_PROTO(void linkblk,(BITS16 *,CELL *,Int));
#endif
#ifdef IDB_USE_MBIT
STATIC_PROTO(CELL *linkcells,(CELL *,Int));
#endif
STATIC_PROTO(Int cmpclls,(CELL *,CELL *,Int));
STATIC_PROTO(Prop GetDBProp,(Atom, unsigned int));
STATIC_PROTO(CELL  CalcKey, (CELL, Term));
STATIC_PROTO(CELL  *MkDBTerm, (CELL *, CELL *, CELL *, CELL *, CELL *, int *));
STATIC_PROTO(DBRef  CreateDBStruct, (Term, DBProp, int));
STATIC_PROTO(DBRef  record, (int, Term, Term, Term));
STATIC_PROTO(DBRef  check_if_cons, (DBRef, Term));
STATIC_PROTO(DBRef  check_if_var, (DBRef));
STATIC_PROTO(DBRef  check_if_wvars, (DBRef, unsigned int, CELL *));
#ifdef IDB_LINK_TABLE
STATIC_PROTO(int  scheckcells, (int, CELL *, CELL *, BITS16 *, CELL));
#endif
STATIC_PROTO(DBRef  check_if_nvars, (DBRef, unsigned int, CELL *));
STATIC_PROTO(Int  p_rcda, (void));
STATIC_PROTO(Int  p_rcdap, (void));
STATIC_PROTO(Int  p_rcdz, (void));
STATIC_PROTO(Int  p_rcdzp, (void));
STATIC_PROTO(Int  p_drcdap, (void));
STATIC_PROTO(Int  p_drcdzp, (void));
STATIC_PROTO(Int  p_rcdaifnot, (void));
STATIC_PROTO(Int  p_rcdzifnot, (void));
STATIC_PROTO(Term  GetDBTerm, (DBRef));
STATIC_PROTO(DBProp  FetchDBPropFromKey, (Term, int, char *));
STATIC_PROTO(Int  i_recorded, (SMALLUNSGN));
STATIC_PROTO(Int  c_recorded, (SMALLUNSGN));
STATIC_PROTO(Int  in_rded, (void));
STATIC_PROTO(Int  co_rded, (void));
STATIC_PROTO(Int  in_rdedp, (void));
STATIC_PROTO(Int  co_rdedp, (void));
STATIC_PROTO(Int  p_first_instance, (void));
STATIC_PROTO(void  ErasePendingRefs, (DBRef));
STATIC_PROTO(void  RemoveDBEntry, (DBRef));
STATIC_PROTO(void  MyEraseClause, (Clause *));
STATIC_PROTO(void  PrepareToEraseClause, (Clause *, DBRef));
STATIC_PROTO(void  EraseEntry, (DBRef));
STATIC_PROTO(Int  p_erase, (void));
STATIC_PROTO(Int  p_eraseall, (void));
STATIC_PROTO(Int  p_erased, (void));
STATIC_PROTO(Int  p_instance, (void));
STATIC_PROTO(int  NotActiveDB, (DBRef));
STATIC_PROTO(DBEntry  *NextDBProp, (PropEntry *));
STATIC_PROTO(Int  init_current_key, (void));
STATIC_PROTO(Int  cont_current_key, (void));
STATIC_PROTO(Int  p_rcdstatp, (void));
STATIC_PROTO(Int  p_somercdedp, (void));
STATIC_PROTO(int  StillInChain, (CODEADDR, PredEntry *));
#ifdef DISCONNECT_OLD_ENTRIES
STATIC_PROTO(yamop * find_next_clause, (DBRef));
STATIC_PROTO(Int  jump_to_next_dynamic_clause, (void));
#endif /* DISCONNECT_OLD_ENTRIES */
#ifdef SFUNC
STATIC_PROTO(void  SFVarIn, (Term));
STATIC_PROTO(void  sf_include, (SFKeep *));
#endif
STATIC_PROTO(Int  p_init_queue, (void));
STATIC_PROTO(Int  p_enqueue, (void));
STATIC_PROTO(void keepdbrefs, (DBRef));
STATIC_PROTO(Int  p_dequeue, (void));
STATIC_PROTO(Int  p_first_age, (void));
STATIC_PROTO(Int  p_db_nb_to_ref, (void));
STATIC_PROTO(Int  p_last_age, (void));
STATIC_PROTO(Int  StealFirstFromDB, (Atom, Int));

#ifdef SUPPORT_HASH_TABLES
/* related property and hint on number of entries */
static void create_hash_table(DBProp p, Int hint) {
  int off = sizeof(CELL)*4, out;
  Int size;

  if (hint < p->NOfEntries)
    hint = p->NOfEntries;
  while (off) {
    int limit = 1 << (off);
    if (inp >= limit) {
      out += off;
      inp >>= off;
    }
    off >>= 1;
  }
  if ((size = 1 << out) < hint)
    hint <<= 1;
  /* clean up the table */
  pt = tbl = (hash_db_entry *)AllocDBSpace(hint*sizeof(hash_db_entry));
  for (i=0; i< hint; i++) {
    pt->key = NULL;
    pt++;
  }
  /* next insert the entries */
}

static void insert_in_table() {
  
}

static void remove_from_table() {
  
}
#endif

#ifdef IDB_LINK_TABLE
inline static CELL *cpcells(CELL *to, CELL *from, Int n)
{
#if HAVE_MEMMOVE
  memmove((void *)to, (void *)from, (size_t)((n+1)*sizeof(CELL)));
  return(to+n+1);
#else
  while (n-- >= 0) {
    *to++ = *from++;
  }
  return(to);
#endif
}

static void linkblk(BITS16 *r,CELL *c,Int n)
{
      register BITS16 p;

      while(n-- > 0) {
	 p = *r++;
	 c[p] = AdjustPtr(c[p], (Int)c);
       }
}
#endif

#ifdef IDB_USE_MBIT
inline static CELL *cpcells(register CELL *to, register CELL *from, Int n)
{
  CELL *last = to + n;
  register CELL off = ((CELL)to)-MBIT;
  while (to <= last) {
    register d0 = *from++;
    if (MARKED(d0)) 
      *to++ = AdjustPtr(d0, off);
    else
      *to++ = d0;
  }
  return(to);
}

static CELL *linkcells(register CELL *to, Int n)
{
  CELL *last = to + n;
  register CELL off = ((CELL)to)-MBIT;
  while(to <= last) {
    register d0 = *to++;
    if (MARKED(d0)) 
      to[-1] = AdjustPtr(d0, off);
  }
  return(to);
}


#endif

static Int cmpclls(CELL *a,CELL *b,Int n)
{
     while (n-- >= 0)
       if(*a++ != *b++) return 0;

     return 1;
}

int DBTrailOverflow()
{
#ifdef IDB_USE_MBIT
  return(FALSE);
#endif
#ifdef IDB_LINK_TABLE
  return((CELL *)lr > (CELL *)TrailTop - 1024);
#endif
}

static Prop 
GetDBProp(Atom ap, unsigned int arity)		/* get DB entry for ap/arity; */
{
  Prop          p0;
  DBProp        p = RepDBProp(p0 = RepAtom(ap)->PropOfAE);
  while (p0 && ((p->KindOfPE != DBProperty) ||
		(p->ArityOfDB != arity) ||
		(p->ModuleOfDB != DBModule))) {
    p = RepDBProp(p0 = p->NextOfPE);
  }
  return (p0);
}



/* These two functions allow us a fast lookup method in the data base */
/* PutMasks builds the mask and hash for a single argument	 */
inline static CELL 
CalcKey(CELL key, Term tw)
{
  /* The first argument is known to be instantiated */
  if (IsAtomTerm(tw)){
    return((AtomHash(tw) | MAtom));
  } else if (IsApplTerm(tw)) {
    Functor f = FunctorOfTerm(tw);
    if (IsExtensionFunctor(f)) {
      if (f == FunctorDouble)
	return(GlobNumHash(tw) | MNum);
      else if (f == FunctorLongInt)
	return(GlobNumHash(tw) | MNum);
      else /* DBRef */
	return(RefHash(tw));	/* Ref */
    } else
      return(FunctorHash(f) | MStruc);
  } else if (IsIntTerm(tw)) {
    return(NumberHash(tw) | MNum);
  } else if (IsPairTerm(tw)) {
    return(FunctorHash(ListFunctor) | MListVar);
  } else {
    DBErrorFlag = OTHER_ERROR_IN_DB;
    DBErrorNumber = TYPE_ERROR_DBTERM;
    DBErrorTerm = tw;
    DBErrorMsg = "in calckey";
    return((CELL)0);
  }
	
}

/* EvalMasks builds the mask and hash for up to three arguments of a term */
CELL 
EvalMasks(register Term tm, CELL *keyp)
{

  if (IsApplTerm(tm)) {
    Functor         fun = FunctorOfTerm(tm);
    unsigned int    arity;

    if (IsExtensionFunctor(fun)) {
      CELL mask, key;

      mask = (FullHashMask << 24) | (FullHashMask << 16) | (FullHashMask << 8) | FullHashMask;
      key = (CELL)fun;
      *keyp = key;
      return(mask);
    }
    else {
      Term tw;
      CELL mask, key;

      arity = ArityOfFunctor(fun);
      key = FullFunctorHash(fun);
      mask = FullHashMask;
#ifdef SFUNC
      if (arity == SFArity) {	/* do not even try to calculate masks */
	*keyp = key;
	return(mask);
      }
#endif
      tw = ArgOfTerm(1, tm);
      if (!IsVarTerm(tw)) {
	mask |= (FullHashMask << 8);
	key |= (CalcKey(key, tw) << 8);
      }
      if (arity == 1) {
	key  |= (((HashMask) << 16) | ((HashMask) << 24));
	mask |= (((FullHashMask) << 16) | ((FullHashMask) << 24));
      } else {
	{
	  Term tw = ArgOfTerm(2, tm);
	  if (!IsVarTerm(tw)) {
	    mask |= (FullHashMask << 16);
	    key |= (CalcKey(key, tw) << 16);
	  }
	}
	if (arity == 2) {
	  key  |= (HashMask << 24);
	  mask |= (FullHashMask << 24);
	} else {
	  Term tw = ArgOfTerm(3, tm);
	  if (!IsVarTerm(tw)) {
	    mask |= (FullHashMask << 24);
	    key |= (CalcKey(key, tw) << 24);
	  }
	}
      }
      *keyp = key;
      return(mask);
    }
  } else if (IsPairTerm(tm)) {
    CELL key  = (FullFunctorHash(ListFunctor) | (HashMask << 24));
    CELL mask = (FullHashMask | (FullHashMask << 24));
    {
      Term tw = HeadOfTerm(tm);
      if (!IsVarTerm(tw)) {
	mask |= (FullHashMask << 8);
	key |= (CalcKey(key, tw) << 8);
      }
    }
    {
      Term tw = TailOfTerm(tm);
      if (!IsVarTerm(tw)) {
	mask |= (FullHashMask << 16);
	key |= (CalcKey(key, tw) << 16);
      }
    }
    *keyp = key;
    return(mask);
  }
  *keyp = 0;
  return(0);
}

/* Called to inform that a new pointer to a data base entry has been added */
#define MarkThisRef(Ref)	((Ref)->NOfRefsTo ++ )

/* From a term, builds its representation in the data base */

#ifdef RATIONAL_TREES
#define COMPRESS_DB_TERM 1
#else
/* #define COMPRESS_DB_TERM 1 */
#endif

#ifdef COMPRESS_DB_TERM
typedef struct {
  CELL *addr;
  CELL val;  
} visitel;
#define UNWIND_CUNIF()                                        \
         while (visited < (visitel *)AuxSp) {                 \
            *(visited->addr) = visited->val;                  \
            visited ++;                                       \
         }
#else
/* otherwise, we just need to restore variables*/
typedef struct {
  CELL *addr;
} visitel;
#define UNWIND_CUNIF()                                        \
         while (visited < (visitel *)AuxSp) {                 \
            RESET_VARIABLE(visited->addr);                    \
            visited ++;                                       \
         }
#endif

/* no checking for overflow while building DB terms yet */
#define  CheckDBOverflow() if (CodeMax+1024 >= (CELL *)visited) {     \
    goto error;					                      \
   }
    
/* no checking for overflow while building DB terms yet */
#define  CheckVisitOverflow() if ((CELL *)to_visit+1024 >= ASP) {     \
    goto error2;					              \
   }
    

/* This routine creates a complex term in the heap. */
static CELL *MkDBTerm(register CELL *pt0, register CELL *pt0_end,
		     register CELL *StoPoint,
		     CELL *CodeMax, CELL *tbase,
		     int *vars_foundp)
{

  register visitel *visited = (visitel *)AuxSp;
  /* store this in H */
  register CELL **to_visit = (CELL **)H;
  CELL **to_visit_base = to_visit;
  /* where we are going to add a new pair */
  int vars_found = 0;

 loop:
  while (pt0 <= pt0_end) {

    CELL *ptd0 = pt0;
    CELL d0 = *ptd0;
  restart:
    if (IsVarTerm(d0))
      goto deref_var; 

    if (IsApplTerm(d0)) {
      register Functor f;
      register CELL *ap2;
      
      /* we will need to link afterwards */
#ifdef IDB_LINK_TABLE
      *lr++ = ToSmall((CELL)(StoPoint)-(CELL)(tbase));
#endif
      ap2 = RepAppl(d0);
      f = (Functor)(*ap2);
      if (IsExtensionFunctor(f)) {
	switch ((Int)f) {
	 case (Int)FunctorDBRef:
	   {
	     DBRef dbentry;
	     /* store now the correct entry */
	     dbentry = DBRefOfTerm(d0);
	     *StoPoint++ = d0;
#ifdef IDB_LINK_TABLE
	     lr--;
#endif
	     dbentry->NOfRefsTo++;
	     *--tofref = dbentry;
	     /* just continue the loop */
	     ++ pt0;
	     continue;
	   }
	case (Int)FunctorDouble:
	  {
	    unsigned int i;
	    ap2++;
#ifdef IDB_USE_MBIT
	    *StoPoint++ = AbsAppl((CELL *)(((CELL)CodeMax-(CELL)tbase)|MBIT));
#else
	    *StoPoint++ = AbsAppl((CELL *)((CELL)CodeMax-(CELL)tbase));
#endif
	    CheckDBOverflow();
	    *CodeMax++ = (CELL)FunctorDouble;
	    for (i=0; i < sizeof(Float)/sizeof(CELL); i++)
	      *CodeMax++ = *ap2++;
	    *CodeMax++ = (CELL)FunctorEndDouble;
	    pt0++;
	    continue;
	  }
	case (Int)FunctorLongInt:
	  {
	    ap2++;
	    CheckDBOverflow();
#ifdef IDB_USE_MBIT
	    *StoPoint++ = AbsAppl((CELL *)(((CELL)CodeMax-(CELL)tbase)|MBIT));
#else
	    *StoPoint++ = AbsAppl((CELL *)((CELL)CodeMax-(CELL)tbase));
#endif
	    *CodeMax++ = (CELL)FunctorLongInt;
	    *CodeMax++ = *ap2++;
	    *CodeMax++ = (CELL)FunctorEndLongInt;
	    pt0++;
	    continue;
	  }
	case (Int)FunctorGeneralRef:
	  {
	    ap2++;
	    CheckDBOverflow();
#ifdef IDB_USE_MBIT
	    *StoPoint++ = AbsAppl((CELL *)(((CELL)CodeMax-(CELL)tbase)|MBIT));
#else
	    *StoPoint++ = AbsAppl((CELL *)((CELL)CodeMax-(CELL)tbase));
#endif
	    *CodeMax++ = (CELL)FunctorGeneralRef;
	    *CodeMax++ = *ap2++;
	    pt0++;
	    continue;
	  }
	default:
	  if (IsAttachFunc(f)) {
	    d0 = attas[ExtFromFunctor(f)].deref_op(d0);
#ifdef IDB_LINK_TABLE
	     lr--;
#endif
	    goto restart;
	  }
	}
      }
#ifdef COMPRESS_DB_TERM
      /* functors are similar to variables!! */
      if (IsNonVarTerm((CELL)f)) {
	/* we've already been here. Just say which value to use */
	*StoPoint = (CELL)f;
#ifdef IDB_LINK_TABLE
	UNMARK(StoPoint);
#endif
	StoPoint++;
	/* just continue the loop */
	++ pt0;
	continue;

      } else {
	/* mark the term as visited by using the main functor */
	visited --;
	visited->addr = ap2;
	visited->val = (CELL)f;
#ifdef IDB_USE_MBIT
	*StoPoint++ = *ap2 = AbsAppl((CELL
				      *)(((CELL)CodeMax-(CELL)tbase)|MBIT));
#else
	*StoPoint++ = *ap2 = AbsAppl((CELL
				      *)((CELL)CodeMax-(CELL)tbase));
#endif /* IDB_USE_MBIT */
#else
	/* first thing, store a link to the list before we move on */
#ifdef IDB_USE_MBIT
	*StoPoint++ = AbsAppl((CELL *)(((CELL)CodeMax-(CELL)tbase)|MBIT));
#else
	*StoPoint++ = AbsAppl((CELL *)((CELL)CodeMax-(CELL)tbase));
#endif
#endif /* COMPRESS_DB_TERM */
	/* next, postpone analysis to the rest of the current list */
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0+1;
	  to_visit[1] = pt0_end;
	  to_visit[2] = StoPoint;
	  to_visit += 3;
	}
	CheckVisitOverflow();
	d0 = ArityOfFunctor(f);
	pt0 = ap2+1;
	pt0_end = ap2 + d0;
	/* prepare for our new compound term */
	/* first the functor */
	CheckDBOverflow();
	*CodeMax++ = (CELL)f;
	/* we'll be working here */
	StoPoint = CodeMax;
	/* now reserve space */
	CodeMax += d0;
	continue;
#ifdef COMPRESS_DB_TERM
      }
#endif
    }
    else if (IsPairTerm(d0)) {
      /* we will need to link afterwards */
#ifdef IDB_LINK_TABLE
      *lr++ = ToSmall((CELL)(StoPoint)-(CELL)(tbase));
#endif
#ifdef COMPRESS_DB_TERM
      if (MARKED(d0)) {
	/* we've already been here. Just say which value to use */
	*StoPoint = d0;
#ifdef IDB_LINK_TABLE
	UNMARK(StoPoint);
#endif
	StoPoint++;
	/* just continue the loop */
	++ pt0;
	continue;

      } else {
	/* now say the compound term has been visited and what was visited */
	visited --;
	visited->addr = pt0;
	visited->val = *pt0;
	/* and store the cell in its value */
#ifdef IDB_USE_MBIT
	*StoPoint++ = *pt0 =
	  AbsPair((CELL *)(((CELL)CodeMax-(CELL)tbase)|MBIT));
#else
	*StoPoint++ = AbsPair((CELL *)((CELL)CodeMax-(CELL)tbase));
	*pt0 = AbsPair((CELL *)(((CELL)CodeMax-(CELL)tbase)|MBIT));
#endif
#else
#ifdef IDB_USE_MBIT
	*StoPoint++ =
	  AbsPair((CELL *)(((CELL)CodeMax-(CELL)tbase)|MBIT));
#else
	*StoPoint++ = AbsPair((CELL *)(((CELL)CodeMax-(CELL)tbase)));
#endif
#endif
	/* next, postpone analysis to the rest of the current list */
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0+1;
	  to_visit[1] = pt0_end;
	  to_visit[2] = StoPoint;
	  to_visit += 3;
	}
	CheckVisitOverflow();
	/* new list */
	/* we are working at CodeMax */
	StoPoint = CodeMax;
	/* set ptr to new term being analysed */
	pt0 = RepPair(d0);
	pt0_end = RepPair(d0) + 1;
	/* reserve space for our new list */
	CodeMax += 2;
	CheckDBOverflow();
	continue;
#ifdef COMPRESS_DB_TERM
      }
#endif
    } else if (IsAtomOrIntTerm(d0)) {
      *StoPoint++ = d0;
      ++pt0;
      continue;
    }
    
    /* the code to dereference a  variable */
  deref_var:
    if (!MARKED(d0))
      {
#if SBA
	if (d0 != 0) {
#else
	if (d0 != (CELL)ptd0) {
#endif
	  ptd0 = (Term *) d0;
	  d0 = *ptd0;
	  goto restart; /* continue dereferencing */
	}
	/* else just drop to found_var */
      }
    /* else just drop to found_var */
    {
      CELL displacement = (CELL)(StoPoint)-(CELL)(tbase);
      
      pt0++;
      /* first time we found this variable! */
      if (!MARKED(d0)) {
	
	/* store previous value */ 
	visited --;
	visited->addr = ptd0;
#ifdef COMPRESS_DB_TERM
	visited->val = d0;
#endif
	CheckDBOverflow();
	/* variables need to be offset at read time */
	*ptd0 = (displacement | MBIT);
#if SBA
	/* the copy we keep will be an empty vaiable   */
	*StoPoint++ = 0;
#else
#ifdef IDB_USE_MBIT
	/* say we've seen the variable, and make it point to its
	   offset */
	/* the copy we keep will be the current displacement   */
	*StoPoint++ = (displacement | MBIT);
#else
	/* the copy we keep will be the current displacement   */
	*StoPoint++ = displacement;
	*lr++ = ToSmall(displacement);
#endif
#endif
	/* indicate we found variables */
	vars_found++;
	continue;
      } else  {
	/* references need to be offset at read time */
#ifdef IDB_LINK_TABLE
	*lr++ = ToSmall(displacement);
#endif
	/* store the offset */
#ifdef IDB_USE_MBIT
	*StoPoint = d0;
#else
	*StoPoint = d0 ^ MBIT;
#endif
	StoPoint++;
	continue;
      }

    }

  }

  /* Do we still have compound terms to visit */
  if (to_visit > (CELL **)to_visit_base) {
    to_visit -= 3;
    pt0 = to_visit[0];
    pt0_end = to_visit[1];
    StoPoint = to_visit[2];
    goto loop;
  }

  /* we're done */
  *vars_foundp = vars_found;
  UNWIND_CUNIF();
  return(CodeMax);
 error:
  DBErrorFlag = OVF_ERROR_IN_DB;
  *vars_foundp = vars_found;
  UNWIND_CUNIF();
  return(NULL);
 error2:
  DBErrorFlag = SOVF_ERROR_IN_DB;
  *vars_foundp = vars_found;
  UNWIND_CUNIF();
  return(NULL);
}


#ifdef SFUNC
/*
 * The sparse terms existing in the structure are to be included now. This
 * means simple copy for constant terms but, some care about variables If
 * they have appeared before, we will know by their position number 
 */
static void 
sf_include(sfp)
	SFKeep         *sfp;
{
  Term            Tm = sfp->SName;
  CELL           *tp = ArgsOfSFTerm(Tm);
  Register Term  *StoPoint = ntp;
  CELL           *displacement = CodeAbs;
  CELL            arg_no;
  Term            tvalue;
  int             j = 3;

  if (sfp->SFather != NIL)
    *(sfp->SFather) = AbsAppl(displacement);
  *StoPoint++ = FunctorOfTerm(Tm);
  *lr++ = ToSmall(displacement + 1);
  *StoPoint++ = (Term) (displacement + 1);
  while (*tp) {
    arg_no = *tp++;
    tvalue = Derefa(tp++);
    if (IsVarTerm(tvalue)) {
      if (((VarKeep *) tvalue)->NOfVars != 0) {
	*StoPoint++ = arg_no;
	*lr++ = ToSmall(displacement + j);
	if (((VarKeep *) tvalue)->New == 0)
	  *StoPoint++ = ((VarKeep *) tvalue)->New = Unsigned(displacement + j);
	else
	  *StoPoint++ = ((VarKeep *) tvalue)->New;
	j += 2;
      }
    } else if (IsAtomicTerm(tvalue)) {
      *StoPoint++ = arg_no;
      *StoPoint++ = tvalue;
      j += 2;
    } else {
      DBErrorFlag = OTHER_ERROR_IN_DB;
      DBErrorNumber = TYPE_ERROR_DBTERM;
      DBErrorTerm = d0;
      DBErrorMsg = "wrong term in SF";
      return(NULL);
    }
  }
  *StoPoint++ = 0;
  ntp = StoPoint;
  CodeAbs = displacement + j;
}
#endif

/*
 * This function is used to check if one of the terms in the idb is the
 * constant to_compare 
 */
inline static DBRef 
check_if_cons(DBRef p, Term to_compare)
{
	while (p != NIL
	 && (p->Flags & (DBCode | ErasedMask | DBVar | DBNoVars | DBComplex)
	     || p->Entry != Unsigned(to_compare)))
		p = NextDBRef(p);
	return (p);
}

/*
 * This function is used to check if one of the terms in the idb is a prolog
 * variable 
 */
static DBRef 
check_if_var(DBRef p)
{
  while (p != NIL &&
	 p->Flags & (DBCode | ErasedMask | DBAtomic | DBNoVars | DBComplex ))
    p = NextDBRef(p);
  return (p);
}

/*
 * This function is used to check if a Prolog complex term with variables
 * already exists in the idb for that key. The comparison is alike ==, but
 * only the relative binding of variables, not their position is used. The
 * comparison is done using the function cmpclls only. The function could
 * only fail if a functor was matched to a Prolog term, but then, it should
 * have failed before because the structure of term would have been very
 * different 
 */
static DBRef 
check_if_wvars(DBRef p, unsigned int NOfCells, CELL *BTptr)
{
  CELL           *memptr;

  do {
    while (p != NIL &&
	   p->Flags & (DBCode | ErasedMask | DBAtomic | DBNoVars | DBVar))
      p = NextDBRef(p);
    if (p == NIL)
      return (p);
    memptr = CellPtr(p->Contents);
    if (NOfCells == p->NOfCells
	&& cmpclls(memptr, BTptr, NOfCells))
      return (p);
    else
      p = NextDBRef(p);
  } while (TRUE);
  return (NIL);
}

#ifdef IDB_LINK_TABLE
static int 
scheckcells(int NOfCells, register CELL *m1, register CELL *m2, BITS16 *lp, register CELL bp)
{
  CELL            base = Unsigned(m1 + 1);
  BITS16         *lp1;

  while (NOfCells-- >= 0) {
    Register CELL   r1, r2;

    r1 = *m1++;
    r2 = *m2++;
    if (r1 == r2)
      continue;
    else if (r2 + bp == r1) {
      /* link pointers may not have been generated in the */
      /* same order */
      /* make sure r1 is really an offset. */
      lp1 = lp;
      r1 = m1 - (CELL *)base;
      while (*lp1 != r1 && *lp1)
	lp1++;
      if (!(*lp1))
	return (FALSE);
      /* keep the old link pointer for future search. */
      /* vsc: this looks like a bug!!!! */
      /* *lp1 = *lp++; */
    } else {
      return (FALSE);
    }
  }
  return (TRUE);
}
#endif

/*
 * the cousin of the previous, but with things a bit more sophisticated.
 * mtchcells, if an error was an found, needs to test ........ 
 */
static DBRef 
check_if_nvars(DBRef p, unsigned int NOfCells, CELL *BTptr)
{
  CELL           *memptr;

  do {
    while (p != NIL &&
	   p->Flags & (DBCode | ErasedMask | DBAtomic | DBComplex | DBVar))
      p = NextDBRef(p);
    if (p == NIL)
      return (p);
    memptr = CellPtr(p->Contents);
#ifdef IDB_LINK_TABLE
    if (scheckcells(NOfCells, memptr, BTptr, LinkAr, Unsigned(p->Contents)))
#else
      if (NOfCells == *memptr++
	  && cmpclls(memptr, BTptr, NOfCells))
#endif
	return (p);
      else
	p = NextDBRef(p);
  } while (TRUE);
  return (NIL);
}

static DBRef 
CreateDBStruct(Term Tm, DBProp p, int InFlag)
{
  Register Term   tt, *nar = NIL;
  Register DBRef  pp0, pp;
  SMALLUNSGN      flag;
  unsigned int    NOfCells = 0;
#ifdef IDB_LINK_TABLE
  int NOfLinks = 0;
#endif
  Term           *ntp0, *ntp;
  /* place DBRefs in ConsultStack */
  DBRef    *TmpRefBase = (DBRef *)ConsultSp;
  CELL	   *CodeAbs;	/* how much code did we find	 */
  int vars_found;

  DBErrorFlag = NO_ERROR_IN_DB;

  if (IsVarTerm(Tm)) {
    Register DBRef  pp;

    tt = Tm;
    if (InFlag & MkIfNot && (found_one = check_if_var(p->First)))
      return (found_one);
    pp = AllocDBSpace(DBLength(NIL));
    if (pp == NIL) {
      DBErrorFlag = OTHER_ERROR_IN_DB;
      DBErrorNumber = SYSTEM_ERROR;
      DBErrorTerm = TermNil;
      DBErrorMsg = "could not allocate space";
      return(NULL);
    }
    pp->id = FunctorDBRef;
    pp->Flags = DBVar;
    pp->Entry = (CELL) Tm;
    return(pp);

  } else if (IsAtomOrIntTerm(Tm)) {
    Register DBRef  pp;
    SMALLUNSGN      flag;


    tt = Tm;
    flag = DBAtomic;
    if (IsAtomOrIntTerm(tt))
    if (InFlag & MkIfNot && (found_one = check_if_cons(p->First, tt)))
      return (found_one);
    pp = AllocDBSpace(DBLength(NIL));
    if (pp == NIL) {
      DBErrorFlag = OTHER_ERROR_IN_DB;
      DBErrorNumber = SYSTEM_ERROR;
      DBErrorTerm = TermNil;
      DBErrorMsg = "could not allocate space";
      return(NULL);
    }
    pp->id = FunctorDBRef;
    pp->Flags = flag;
    pp->Entry = (CELL) Tm;
    return(pp);

  } else {

    tofref = (DBRef *)TmpRefBase;
    /* compound term */
    pp0 = (DBRef)PreAllocCodeSpace();
    ntp0 = pp0->Contents;
#ifdef IDB_LINK_TABLE
    lr = LinkAr = (BITS16 *)TR;
#endif
    if (IsPairTerm(Tm)) {
      /* avoid null pointers!! */
      tt = AbsPair((CELL *)sizeof(CELL));
      ntp = MkDBTerm(RepPair(Tm), RepPair(Tm)+1, ntp0, ntp0+2, ntp0, &vars_found);
      if (ntp == NULL) {
	return(NULL);
      }
    }
    else
    {
      unsigned int arity;
      Functor fun;

      tt = AbsAppl((CELL *)sizeof(CELL));
      /* we need to store the functor manually */
      fun = (Functor)(*ntp0 = (CELL)FunctorOfTerm(Tm));
      if (fun == FunctorDouble) {
	unsigned int i;
	CELL *fp = RepAppl(Tm)+1;

	ntp = ntp0+1;
	for (i=0; i < sizeof(Float)/sizeof(CELL); i++) {
	  *ntp++ = *fp++;
	}
	*ntp++ = (CELL)FunctorEndDouble;
      } else if (fun == FunctorDBRef) {
	DBRef dbr;

	pp = AllocDBSpace(DBLength(2*sizeof(DBRef)));
	if (pp == NIL) {
	  DBErrorFlag = OTHER_ERROR_IN_DB;
	  DBErrorNumber = SYSTEM_ERROR;
	  DBErrorTerm = TermNil;
	  DBErrorMsg = "could not allocate space";
	  return(NULL);
	}
	pp->id = FunctorDBRef;
	pp->Flags = DBNoVars|DBComplex|DBWithRefs;
	pp->Entry = Tm;
	pp->NOfCells = 2;
	dbr = DBRefOfTerm(Tm);
	dbr->NOfRefsTo++;
	pp->Contents[0] = (CELL)NIL;
	pp->Contents[1] = (CELL)dbr;
	pp->Code = (CODEADDR)(pp->Contents+2);
	return(pp);
      } else if (fun == FunctorLongInt) {
	ntp = ntp0+1;
	*ntp++ = (RepAppl(Tm))[1];
	*ntp++ = (CELL)FunctorEndLongInt;
      } else {
	arity = ArityOfFunctor(fun);
	ntp = MkDBTerm(RepAppl(Tm)+1,
		     RepAppl(Tm)+arity,
		     ntp0+1, ntp0+1+arity, ntp0, &vars_found);
	if (ntp == NULL)
	  return(NULL);
      }
    } 
    CodeAbs = (CELL *)((CELL)ntp-(CELL)ntp0);
    if (DBErrorFlag)
      return (NULL);	/* Error Situation */
    NOfCells = (ntp - 1) - ntp0;	/* End Of Code Info */
#ifdef IDB_LINK_TABLE
    NOfLinks = (lr - LinkAr);
#endif
    if (vars_found || InFlag & InQueue) {
      /*
       * Take into account the fact that one needs an entry
       * for the number of links 
       */
      flag = DBComplex;
#ifdef IDB_LINK_TABLE
      CodeAbs++;	/* We have one more cell */
      lr++;
      /*
       * One adds one plus in order to account for the
       * differences in sizes 
       */
      CodeAbs += CellPtr(lr) - CellPtr(LinkAr) + 1;
      if ((CELL *)((char *)ntp0+(CELL)CodeAbs) > AuxSp) {
	DBErrorFlag = OVF_ERROR_IN_DB;
	return(NULL);
      }
      /* restore lr to NULL in case there is a TR overflow */
      lr = NULL;
#endif
      if ((InFlag & MkIfNot) && (found_one = check_if_wvars(p->First, NOfCells, ntp0)))
	return (found_one);
    } else {
#ifdef IDB_LINK_TABLE
      /* make sure lr ends in 0 for check_if_nvars */  
      *lr++ = 0;
      /* restore lr to NULL in case there is a TR overflow */
      lr = NULL;
#endif
      flag = DBNoVars;
      if ((InFlag & MkIfNot) && (found_one = check_if_nvars(p->First, NOfCells, ntp0)))
	return (found_one);
    }
    if (tofref != TmpRefBase) {
      CodeAbs += TmpRefBase - tofref + 1;
      if ((CELL *)((char *)ntp0+(CELL)CodeAbs) > AuxSp) {
	DBErrorFlag = OVF_ERROR_IN_DB;
	return(NULL);
      }
      flag |= DBWithRefs;
    }
#ifdef IDB_LINK_TABLE
    if (Unsigned(CodeAbs) >= 0x40000) {
      DBErrorFlag = OTHER_ERROR_IN_DB;
      DBErrorNumber = SYSTEM_ERROR;
      DBErrorTerm = TermNil;
      DBErrorMsg = "trying to store term larger than 256K";
      return(NULL);
    }
#endif
    pp = AllocDBSpace(DBLength(CodeAbs));
    if (pp == NIL) {
      DBErrorFlag = OVF_ERROR_IN_DB;
      DBErrorNumber = SYSTEM_ERROR;
      DBErrorTerm = TermNil;
      DBErrorMsg = "heap crashed against stacks";
      return(NULL);
    }
    pp->id = FunctorDBRef;
    pp->Flags = flag;
    if (flag & DBComplex) {
#ifdef IDB_LINK_TABLE
      BITS16         *woar;
#endif /* IDB_LINK_TABLE */

      pp->NOfCells = NOfCells;
      if (pp0 != pp) {
	nar = pp->Contents;
#ifdef IDB_LINK_TABLE
	nar = (Term *) cpcells(CellPtr(nar), ntp0, Unsigned(NOfCells));
#endif
#ifdef IDB_USE_MBIT
	memcpy((void *)nar, (const void *)ntp0,
	       (size_t)((NOfCells+1)*sizeof(CELL)));
	nar += NOfCells+1;
#endif
      } else {
	nar = pp->Contents + Unsigned(NOfCells)+1;
      }
#ifdef IDB_LINK_TABLE
      woar = WordPtr(nar);
      *woar++ = NOfLinks;
      memcpy((void *)woar,(const void *)LinkAr,(size_t)(NOfLinks*sizeof(BITS16)));
      woar += NOfLinks;
#ifdef ALIGN_LONGS
#if SIZEOF_INT_P==8
      while ((Unsigned(woar) & 7) != 0)
	woar++;		
#else
      if ((Unsigned(woar) & 3) != 0)
	woar++;
#endif
#endif
      nar = (Term *) (woar);
#endif
      pp->Entry = (CELL) tt;
    } else if (flag & DBNoVars) {
      if (pp0 != pp) {
	nar = (Term *) cpcells(CellPtr(pp->Contents), ntp0, Unsigned(NOfCells));
      } else {
#ifdef IDB_LINK_TABLE
	nar = pp->Contents + Unsigned(NOfCells)+1;
#endif
#ifdef IDB_USE_MBIT
	/* we still need to link */
	nar = (Term *) linkcells(ntp0, NOfCells);
#endif
      }
#ifdef IDB_LINK_TABLE
      linkblk(LinkAr, CellPtr(pp->Contents), Unsigned(NOfLinks));
#endif
      pp->Entry = AdjustPtr(tt,Unsigned(pp->Contents)-sizeof(CELL));
      pp->NOfCells = NOfCells;
    }
    if (flag & DBWithRefs) {
      DBRef *ptr = TmpRefBase;

      *nar++ = 0;
      while (ptr != tofref)
	*nar++ = (CELL)(*--ptr);
      pp->Code = CodePtr(nar);

    } else {

      pp->Code = 0;

    }
    return (pp);
  }
}

static DBRef 
record(int Flag, Term key, Term t_data, Term t_code)
{
  Register Term   twork = key;
  Register DBProp p;
  Register DBRef  x;


#ifdef SFUNC
  FathersPlace = NIL;
#endif
  if (EndOfPAEntr(p = FetchDBPropFromKey(twork, TRUE, "record/3"))) {
    return(NULL);
  }
  YAPEnterCriticalSection();
  if ((x = CreateDBStruct(t_data, p, Flag)) == NULL) {
    YAPLeaveCriticalSection();
    return (NULL);
  }
  if ((Flag & MkIfNot) && found_one)
    return (NULL);
  TRAIL_REF(&(x->Flags));
  if (x->Flags & (DBNoVars|DBComplex))
    x->Mask = EvalMasks(t_data, &x->Key);
  else
    x->Mask = x->Key = 0;
  if (Flag & MkCode)
    x->Flags |= DBCode;
  else
    x->Flags |= DBNoCode;
  x->Parent = ParentCast(p);
  x->Flags |= (InUseMask | DBClMask);
  x->NOfRefsTo = 0;
#ifdef KEEP_ENTRY_AGE
  if (Flag & MkFirst)
    x->age = -(p->age++);
  else
    x->age = (p->age++);
#endif /* KEEP_ENTRY_AGE */
  if (p->First == NIL) {
    p->First = p->Last = x;
#ifdef KEEP_OLD_ENTRIES_HANGING_ABOUT
    p->FirstNEr = x;
#endif
    x->Prev = x->Next = NIL;
  } else if (Flag & MkFirst) {
    x->Prev = NIL;
    (p->First)->Prev = x;
    x->Next = p->First;
    p->First = x;
#ifdef KEEP_OLD_ENTRIES_HANGING_ABOUT
    p->FirstNEr = x;
#endif
  } else {
    x->Next = NIL;
    (p->Last)->Next = x;
    x->Prev = p->Last;
    p->Last = x;
#ifdef KEEP_OLD_ENTRIES_HANGING_ABOUT
    if (p->FirstNEr == NIL)
      p->FirstNEr = x;
#endif
  }
  if (Flag & WithRef) {
    x->Code = (CODEADDR) GeneralRefOfTerm(t_code);
  }
  YAPLeaveCriticalSection();
  return (x);
}

/* recorda(+Functor,+Term,-Ref) */
static Int 
p_rcda(void)
{
  /* Idiotic xlc's cpp does not work with ARG1 within MkDBRefTerm */
  Term            TRef, t1 = Deref(ARG1), t2 = Deref(ARG2);

  DBModule = 0;
  if (!IsVarTerm(Deref(ARG3)))
    return (FALSE);
 restart_record:
  TRef = MkDBRefTerm(record(MkFirst, t1, t2, Unsigned(0)));
  switch(DBErrorFlag) {
  case NO_ERROR_IN_DB:
    return (unify(ARG3, TRef));
  case SOVF_ERROR_IN_DB:
    if (!gc(3, ENV, P)) {
      Abort("[ SYSTEM ERROR: YAP could not grow stack in recorda/3 ]\n");
      return(FALSE);
    }
    goto recover_record;
  case OVF_ERROR_IN_DB:
    if (!growheap()) {
      Abort("[ SYSTEM ERROR: YAP could not grow heap in recorda/3 ]\n");
      return(FALSE);
    } else
      goto recover_record;
  default:
    Error(DBErrorNumber, DBErrorTerm, DBErrorMsg);
    return(FALSE);
  }
 recover_record:
  DBErrorFlag = NO_ERROR_IN_DB;
  t1 = Deref(ARG1);
  t2 = Deref(ARG2);
  goto restart_record;
}

/* '$recordap'(+Functor,+Term,-Ref) */
static Int 
p_rcdap(void)
{
  Term            TRef, t1 = Deref(ARG1), t2 = Deref(ARG2);

  DBModule = CurrentModule;
  if (!IsVarTerm(Deref(ARG3)))
    return (FALSE);
 restart_record:
  TRef = MkDBRefTerm(record(MkFirst | MkCode, t1, t2, Unsigned(0)));
  switch(DBErrorFlag) {
  case NO_ERROR_IN_DB:
    return (unify(ARG3, TRef));
  case SOVF_ERROR_IN_DB:
    if (!gc(3, ENV, P)) {
      Abort("[ SYSTEM ERROR: YAP could not grow stack in recorda/3 ]\n");
      return(FALSE);
    }
    goto recover_record;
  case OVF_ERROR_IN_DB:
    if (!growheap()) {
      Abort("[ SYSTEM ERROR: YAP could not grow heap in recorda/3 ]\n");
      return(FALSE);
    } else
      goto recover_record;
  default:
    Error(DBErrorNumber, DBErrorTerm, DBErrorMsg);
    return(FALSE);
  }
 recover_record:
  DBErrorFlag = NO_ERROR_IN_DB;
  t1 = Deref(ARG1);
  t2 = Deref(ARG2);
  goto restart_record;
}

/* recordz(+Functor,+Term,-Ref) */
static Int 
p_rcdz(void)
{
  Term            TRef, t1 = Deref(ARG1), t2 = Deref(ARG2);

  DBModule = 0;
  if (!IsVarTerm(Deref(ARG3)))
    return (FALSE);
 restart_record:
  TRef = MkDBRefTerm(record(MkLast, t1, t2, Unsigned(0)));
  switch(DBErrorFlag) {
  case NO_ERROR_IN_DB:
    return (unify(ARG3, TRef));
  case SOVF_ERROR_IN_DB:
    if (!gc(3, ENV, P)) {
      Abort("[ SYSTEM ERROR: YAP could not grow stack in recordz/3 ]\n");
      return(FALSE);
    }
    goto recover_record;
  case OVF_ERROR_IN_DB:
    if (!growheap()) {
      Abort("[ SYSTEM ERROR: YAP could not grow heap in recordz/3 ]\n");
      return(FALSE);
    } else
      goto recover_record;
  default:
    Error(DBErrorNumber, DBErrorTerm, DBErrorMsg);
    return(FALSE);
  }
 recover_record:
  DBErrorFlag = NO_ERROR_IN_DB;
  t1 = Deref(ARG1);
  t2 = Deref(ARG2);
  goto restart_record;
}

/* '$recordzp'(+Functor,+Term,-Ref) */
static Int 
p_rcdzp(void)
{
  Term            TRef, t1 = Deref(ARG1), t2 = Deref(ARG2);

  DBModule = CurrentModule;
  if (!IsVarTerm(Deref(ARG3)))
    return (FALSE);
 restart_record:
  TRef = MkDBRefTerm(record(MkLast | MkCode, t1, t2, Unsigned(0)));
  switch(DBErrorFlag) {
  case NO_ERROR_IN_DB:
    return (unify(ARG3, TRef));
  case SOVF_ERROR_IN_DB:
    if (!gc(3, ENV, P)) {
      Abort("[ SYSTEM ERROR: YAP could not grow stack in recordz/3 ]\n");
      return(FALSE);
    }
    goto recover_record;
  case OVF_ERROR_IN_DB:
    if (!growheap()) {
      Abort("[ SYSTEM ERROR: YAP could not grow heap in recordz/3 ]\n");
      return(FALSE);
    } else
      goto recover_record;
  default:
    Error(DBErrorNumber, DBErrorTerm, DBErrorMsg);
    return(FALSE);
  }
 recover_record:
  DBErrorFlag = NO_ERROR_IN_DB;
  t1 = Deref(ARG1);
  t2 = Deref(ARG2);
  goto restart_record;
}

/* '$record_stat_source'(+Functor,+Term) */
static Int 
p_rcdstatp(void)
{
  Term t1 = Deref(ARG1), t2 = Deref(ARG2), t3 = Deref(ARG3);
  int mk_first;

  DBModule = CurrentModule;
  if (IsVarTerm(t3) || !IsIntTerm(t3))
    return (FALSE);
  if (IsVarTerm(t3) || !IsIntTerm(t3))
    return (FALSE);
  mk_first = ((IntOfTerm(t3) % 4) == 2);
 restart_record:
  if (mk_first)
    (void)record(MkFirst | MkCode, t1, t2, Unsigned(0));
  else
    (void) record(MkLast | MkCode, t1, t2, Unsigned(0));
  switch(DBErrorFlag) {
  case NO_ERROR_IN_DB:
    return (TRUE);
  case SOVF_ERROR_IN_DB:
    if (!gc(3, ENV, P)) {
      Abort("[ SYSTEM ERROR: YAP could not grow stack in recorda/3 ]\n");
      return(FALSE);
    }
    goto recover_record;
  case OVF_ERROR_IN_DB:
    if (!growheap()) {
      Abort("[ SYSTEM ERROR: YAP could not grow heap in recorda/3 ]\n");
      return(FALSE);
    } else
      goto recover_record;
  default:
    Error(DBErrorNumber, DBErrorTerm, DBErrorMsg);
    return(FALSE);
  }
 recover_record:
  DBErrorFlag = NO_ERROR_IN_DB;
  t1 = Deref(ARG1);
  t2 = Deref(ARG2);
  goto restart_record;
}

/* '$recordap'(+Functor,+Term,-Ref,+CRef) */
static Int 
p_drcdap(void)
{
  Term            TRef, t1 = Deref(ARG1), t2 = Deref(ARG2), t4 = Deref(ARG4);

  DBModule = CurrentModule;
  if (!IsVarTerm(Deref(ARG3)))
    return (FALSE);
  if (IsVarTerm(t4) || !IsGeneralRefTerm(t4))
    return (FALSE);
 restart_record:
  TRef = MkDBRefTerm(record(MkFirst | MkCode | WithRef,
			    t1, t2, t4));
  switch(DBErrorFlag) {
  case NO_ERROR_IN_DB:
    return (unify(ARG3, TRef));
  case SOVF_ERROR_IN_DB:
    if (!gc(4, ENV, P)) {
      Abort("[ SYSTEM ERROR: YAP could not grow stack in recorda/3 ]\n");
      return(FALSE);
    }
    goto recover_record;
  case OVF_ERROR_IN_DB:
    if (!growheap()) {
      Abort("[ SYSTEM ERROR: YAP could not grow heap in recorda/3 ]\n");
      return(FALSE);
    } else
      goto recover_record;
  default:
    Error(DBErrorNumber, DBErrorTerm, DBErrorMsg);
    return(FALSE);
  }
 recover_record:
  DBErrorFlag = NO_ERROR_IN_DB;
  t1 = Deref(ARG1);
  t2 = Deref(ARG2);
  t4 = Deref(ARG4);
  goto restart_record;
}

/* '$recordzp'(+Functor,+Term,-Ref,+CRef) */
static Int 
p_drcdzp(void)
{
  Term            TRef, t1 = Deref(ARG1), t2 = Deref(ARG2), t4 =  Deref(ARG4);

  DBModule = CurrentModule;
  if (!IsVarTerm(Deref(ARG3)))
    return (FALSE);
  if (IsVarTerm(t4) || !IsGeneralRefTerm(t4))
    return (FALSE);
 restart_record:
  TRef = MkDBRefTerm(record(MkLast | MkCode | WithRef,
			    t1, t2, t4));
  switch(DBErrorFlag) {
  case NO_ERROR_IN_DB:
    return (unify(ARG3, TRef));
  case SOVF_ERROR_IN_DB:
    if (!gc(4, ENV, P)) {
      Abort("[ SYSTEM ERROR: YAP could not grow stack in recordz/3 ]\n");
      return(FALSE);
    }
    goto recover_record;
  case OVF_ERROR_IN_DB:
    if (!growheap()) {
      Abort("[ SYSTEM ERROR: YAP could not grow heap in recordz/3 ]\n");
      return(FALSE);
    } else
      goto recover_record;
  default:
    Error(DBErrorNumber, DBErrorTerm, DBErrorMsg);
    return(FALSE);
  }
 recover_record:
  DBErrorFlag = NO_ERROR_IN_DB;
  t1 = Deref(ARG1);
  t2 = Deref(ARG2);
  t4 = Deref(ARG4);
  goto restart_record;
}

/* '$recordaifnot'(+Functor,+Term,-Ref) */
static Int 
p_rcdaifnot(void)
{
  Term            TRef;
  DBRef           db_ref;

 restart_record:
  DBModule = 0;
  if (!IsVarTerm(Deref(ARG3)))
    return (FALSE);
  found_one = NIL;
  db_ref = record(MkFirst | MkIfNot, Deref(ARG1), Deref(ARG2), Unsigned(0));
  if (db_ref == NULL)
    return(FALSE);
  switch(DBErrorFlag) {
  case NO_ERROR_IN_DB:
    TRef = MkDBRefTerm(db_ref);
    return (unify(ARG3, TRef));
  case SOVF_ERROR_IN_DB:
    if (!gc(3, ENV, P)) {
      Abort("[ SYSTEM ERROR: YAP could not grow stack in recordaifnot/3 ]\n");
      return(FALSE);
    }
    goto recover_record;
  case OVF_ERROR_IN_DB:
    if (!growheap()) {
      Abort("[ SYSTEM ERROR: YAP could not grow heap in recordzifnot/3 ]\n");
      return(FALSE);
    } else
      goto recover_record;
  default:
    Error(DBErrorNumber, DBErrorTerm, DBErrorMsg);
    return(FALSE);
  }
 recover_record:
  DBErrorFlag = NO_ERROR_IN_DB;
  goto restart_record;
}

/* '$recordzifnot'(+Functor,+Term,-Ref) */
static Int 
p_rcdzifnot(void)
{
  Term            TRef;
  DBRef           db_ref;

 restart_record:
  DBModule = 0;
  if (!IsVarTerm(Deref(ARG3)))
    return (FALSE);
  found_one = NIL;
  db_ref = record(MkLast | MkIfNot, Deref(ARG1), Deref(ARG2), Unsigned(0));
  if (db_ref == NULL)
    return(FALSE);
  switch(DBErrorFlag) {
  case NO_ERROR_IN_DB:
    TRef = MkDBRefTerm(db_ref);
    return (unify(ARG3, TRef));
  case SOVF_ERROR_IN_DB:
    if (!gc(3, ENV, P)) {
      Abort("[ SYSTEM ERROR: YAP could not grow stack in recordaifnot/3 ]\n");
      return(FALSE);
    }
    goto recover_record;
  case OVF_ERROR_IN_DB:
    if (!growheap()) {
      Abort("[ SYSTEM ERROR: YAP could not grow heap in recordzifnot/3 ]\n");
      return(FALSE);
    } else
      goto recover_record;
  default:
    Error(DBErrorNumber, DBErrorTerm, DBErrorMsg);
    return(FALSE);
  }
 recover_record:
  DBErrorFlag = NO_ERROR_IN_DB;
  goto restart_record;
}

static Term 
GetDBTerm(DBRef DBSP)
{
  if (DBSP->Flags & (DBNoVars | DBAtomic))
    return ((Term) DBSP->Entry);
  if (DBSP->Flags & DBComplex) {
    CELL           *HOld = H;
    CELL           *HeapPtr = H;
    CELL           *pt;
    CELL            NOf;

    pt = CellPtr(DBSP->Contents);
    NOf = DBSP->NOfCells;
    if (H+NOf > ASP - MinStackGap*(stack_overflows+1)) {
      return((Term)0);
    }
    HeapPtr = cpcells(HeapPtr, pt, NOf);
    pt += HeapPtr - H;
#ifdef IDB_LINK_TABLE
    {
      BITS16 *lp = (BITS16 *)pt;
      NOf = *lp++;
      linkblk(lp, CellPtr(HOld), NOf);
    }
#endif
    H = HeapPtr;
    return (AdjustPtr((Term)(DBSP->Entry),Unsigned(HOld)-sizeof(CELL)));
  }
  return (MkVarTerm());
}


static DBProp
FetchDBPropFromKey(Term twork, int new, char *error_mssg)
{
  Atom At;
  Int arity;

  if (IsVarTerm(twork)) {
    Error(INSTANTIATION_ERROR, twork, error_mssg);
    return(RepDBProp(NIL));
  } else if (IsApplTerm(twork)) {
    Register Functor f = FunctorOfTerm(twork);
    if (IsExtensionFunctor(f)) {
      Error(TYPE_ERROR_KEY, twork, error_mssg);
      return(RepDBProp(NIL));
    }
    At = NameOfFunctor(f);
    arity = ArityOfFunctor(f);
  } else if (IsAtomTerm(twork)) {
    arity = 0, At = AtomOfTerm(twork);
  } else if (IsPairTerm(twork)) {
    At = AtomDot;
    arity = 2;
  } else if (IsIntTerm(twork)) {
    char            s[20];
#if SHORT_INTS
    sprintf(s, "%ld", IntOfTerm(twork));
#else
    sprintf(s, "%d", IntOfTerm(twork));
#endif
    At = LookupAtom(s);
    arity = 0;
  } else {
    Error(TYPE_ERROR_KEY, twork,error_mssg);
    return(RepDBProp(NIL));
  }
  if (new) {
    DBProp p;

    if (EndOfPAEntr(p = RepDBProp(GetDBProp(At, arity)))) {
     /* create a new DBProp				 */
      p = (DBProp) AllocAtomSpace(sizeof(*p));
      p->KindOfPE = DBProperty;
      p->ArityOfDB = arity;
      p->First = p->Last = NIL;
#ifdef KEEP_OLD_ENTRIES_HANGING_ABOUT
      p->FirstNEr = NIL;
#endif
      p->ModuleOfDB = DBModule;
      /* This is NOT standard but is QUITE convenient */
      if (arity == 0)
	p->FunctorOfDB = (Functor) At;
      else
	p->FunctorOfDB = MkFunctor(At,arity);
#ifdef KEEP_ENTRY_AGE
      p->age = 0;
#endif /* KEEP_ENTRY_AGE */
      p->NextOfPE = RepAtom(At)->PropOfAE;
      RepAtom(At)->PropOfAE = AbsDBProp(p);
    }
    return(p);
  } else
    return(RepDBProp(FindDBProp(At, arity)));
}

/* Finds a term recorded under the key ARG1			 */
static Int 
i_recorded(SMALLUNSGN flagsToAvoid)
{
  Term            TermDB, TRef;
  Register DBRef  ref;
  DBProp          AtProp;
  CELL           *PreviousHeap;
  CELL            mask = 0, key = 0;
  Register Term   twork = Deref(ARG1);	/* initially working with
					 * ARG1 */
  register choiceptr b0=B;

  Term t3 = Deref(ARG3);
  if (!IsVarTerm(t3) && !IsDBRefTerm(t3)) {
    cut_fail();
  }
  if (EndOfPAEntr(AtProp = FetchDBPropFromKey(twork, FALSE, "recorded/3"))) {
    if (b0 == B)
      cut_fail();
    else
      return(FALSE);
  }
#ifdef KEEP_OLD_ENTRIES_HANGING_ABOUT
  ref = AtProp->FirstNEr;
#else
  ref = AtProp->First;
#endif
  while (ref != NIL
	 && (ref->Flags & flagsToAvoid))
    ref = NextDBRef(ref);
  if (ref == NIL) {
    cut_fail();
  }
  twork = Deref(ARG2);	/* now working with ARG2 */
  if (IsVarTerm(twork)) {
    mask = key = 0;
    EXTRA_CBACK_ARG(3,2) = MkIntegerTerm(mask);
    EXTRA_CBACK_ARG(3,3) = MkIntegerTerm(key);
    B->cp_h = H;
    while ((TermDB = GetDBTerm(ref)) == (CELL)0) {
      /* oops, we are in trouble, not enough stack space */
      gc(3, ENV, CP);
      PreviousHeap = H;
      twork = Deref(ARG2);
      t3 = Deref(ARG3);
    }
    if (!unify(twork, TermDB)) {
      cut_fail();
    }
  } else if (IsAtomOrIntTerm(twork)) {
    mask = 0;
    key = Unsigned(twork);
    EXTRA_CBACK_ARG(3,2) = MkIntegerTerm(mask);
    EXTRA_CBACK_ARG(3,3) = MkIntegerTerm(key);
    B->cp_h = H;
    do {
      if (((twork == ref->Entry) || (ref->Flags & DBVar)) &&
	  !(ref->Flags & flagsToAvoid))
	break;
      ref = NextDBRef(ref);
      if (ref == NIL) {
	cut_fail();
      }
    } while (TRUE);
  } else {
    mask = EvalMasks(twork, &key);
    EXTRA_CBACK_ARG(3,2) = MkIntegerTerm(((Int)mask));
    EXTRA_CBACK_ARG(3,3) = MkIntegerTerm(((Int)key));
    PreviousHeap = H;
    B->cp_h = H;
    do {
      H = PreviousHeap;
      while ((mask & ref->Key) != (key & ref->Mask)) {
	while ((ref = NextDBRef(ref)) != NIL
	       && (ref->Flags & flagsToAvoid));
	if (ref == NIL) {
	  cut_fail();
	}
      }
      while ((TermDB = GetDBTerm(ref)) == (CELL)0) {
	/* oops, we are in trouble, not enough stack space */
	gc(3, ENV, CP);
	PreviousHeap = H;
      }
      if (unify(ARG2, TermDB))
	break;
      while ((ref = NextDBRef(ref)) != NIL
	     && (ref->Flags & flagsToAvoid));
      if (ref == NIL) {
	cut_fail();
      }
    } while (TRUE);
  }
  EXTRA_CBACK_ARG(3,1) = (CELL)ref;
  /* This should be after any non-tagged terms, because the routines in grow.c
     go from upper to lower addresses */
  TRef = MkDBRefTerm(ref);
  if (!(ref->Flags & InUseMask)) {
    ref->Flags |= InUseMask;
    TRAIL_REF(&(ref->Flags));	/* So that fail will erase it */
  }
  return (unify(ARG3, TRef));
}

static Int 
c_recorded(SMALLUNSGN flagsToAvoid)
{
  Term            TermDB, TRef;
  Register DBRef  ref, ref0;
  CELL           *PreviousHeap = H;
  CELL            mask, key;

  ref = NextDBRef(ref0 = (DBRef)EXTRA_CBACK_ARG(3,1));
  if (ref == NIL) {
#ifdef DISCONNECT_OLD_ENTRIES
    if (ref0->Flags & ErasedMask) {
      Int my_age = ref0->age;
      /* we were thrown out of the hash chain */
      ref = ((DBProp)(ref0->Parent))->First;
      /* search for an old entry */
      while (ref != NIL && ref->age < my_age)
	ref = ref->Next;
      /* we have used the DB entry, so we can remove it now, although
	 first we have to make sure noone is pointing to it */
      if (!(ref0->Flags & InUseMask) && (ref0->NOfRefsTo == 0)) {
	/* I can't free space for a clause if it's still being pointed
	   to from code */
	if ((ref0->Flags & DBCode) && ref0->Code) {
	  Clause *clau = ClauseCodeToClause(ref0->Code);
	  if (!(clau->ClFlags & InUseMask)) {
	    FreeDBSpace((char *) ref0);
	  }
	} else {
	  FreeDBSpace((char *) ref0);
	}
      }
      if (ref == NIL) {
	cut_fail();
      }
    }
    else
#endif
      cut_fail();
  }
	
  {
    Term ttmp = EXTRA_CBACK_ARG(3,2);
    if (IsLongIntTerm(ttmp))
      mask = (CELL)LongIntOfTerm(ttmp);
    else
      mask = (CELL)IntOfTerm(ttmp);
  }
  {
    Term ttmp = EXTRA_CBACK_ARG(3,3);
    if (IsLongIntTerm(ttmp))
      key = (CELL)LongIntOfTerm(ttmp);
    else
      key = (CELL)IntOfTerm(ttmp);
  }
  while (ref != NIL
	 && (ref->Flags & flagsToAvoid))
    ref = NextDBRef(ref);
  if (ref == NIL) {
    cut_fail();
  }
  if (mask == 0 && key == 0) {	/* ARG2 is a variable */
    while ((TermDB = GetDBTerm(ref)) == (CELL)0) {
      /* oops, we are in trouble, not enough stack space */
      gc(3, ENV, CP);
      PreviousHeap = H;
    }
    unify(ARG2, TermDB);
  } else if (mask == 0) {	/* ARG2 is a constant */
    do {
      if (((key == Unsigned(ref->Entry)) || (ref->Flags & DBVar)) &&
	  !(ref->Flags & flagsToAvoid))
	break;
      ref = NextDBRef(ref);
    } while (ref != NIL);
    if (ref == NIL) {
      cut_fail();
    }
  } else
    do {		/* ARG2 is a structure */
      H = PreviousHeap;
      while ((mask & ref->Key) != (key & ref->Mask)) {
	while ((ref = NextDBRef(ref)) != NIL
	       && (ref->Flags & flagsToAvoid));
	if (ref == NIL) {
	  cut_fail();
	}
      }
      while ((TermDB = GetDBTerm(ref)) == (CELL)0) {
	/* oops, we are in trouble, not enough stack space */
	gc(3, ENV, CP);
	PreviousHeap = H;
      }
      if (unify(ARG2, TermDB))
	break;
      while ((ref = NextDBRef(ref)) != NIL
	     && (ref->Flags & flagsToAvoid));
      if (ref == NIL) {
	cut_fail();
      }
    } while (1);
  TRef = MkDBRefTerm(ref);
  EXTRA_CBACK_ARG(3,1) = (CELL)ref;
  if (!(ref->Flags & InUseMask)) {
    ref->Flags |= InUseMask;
    TRAIL_REF(&(ref->Flags));	/* So that fail will erase it */
  }
  return (unify(ARG3, TRef));
}

/*
 * The arguments for this 4 functions are the flags for terms which should be
 * skipped 
 */

/* recorded(+Functor,+Term,-Ref) */
static Int 
in_rded(void)
{
  DBModule = 0;
  return (i_recorded((SMALLUNSGN) (DBCode | ErasedMask)));
}

static Int 
co_rded(void)
{
  DBModule = 0;
  return (c_recorded((SMALLUNSGN) (DBCode | ErasedMask)));
}

/* '$recordedp'(+Functor,+Term,-Ref) */
static Int 
in_rdedp(void)
{
  DBModule = CurrentModule;
  return (i_recorded((SMALLUNSGN) (DBNoCode | ErasedMask)));
}


static Int 
co_rdedp(void)
{
  DBModule = CurrentModule;
  return (c_recorded((SMALLUNSGN) (DBNoCode | ErasedMask)));
}

/* '$some_recordedp'(Functor)				 */
static Int 
p_somercdedp(void)
{
  Register DBRef  ref;
  DBProp            AtProp;
  Register Term   twork = Deref(ARG1);	/* initially working with
						 * ARG1 */
  DBModule = CurrentModule;
  if (EndOfPAEntr(AtProp = FetchDBPropFromKey(twork, FALSE, "some_recorded/3"))) {
    return(FALSE);
  }
  ref = FrstDBRef(AtProp);
  while (ref != NIL && (ref->Flags & (DBNoCode | ErasedMask)))
    ref = NextDBRef(ref);
  if (ref == NIL)
    return (FALSE);
  else
    return (TRUE);
}

/* Finds the first instance recorded under key ARG1			 */
static Int 
p_first_instance(void)
{
  Term            TRef;
  Register DBRef  ref;
  DBProp          AtProp;
  Register Term   twork = Deref(ARG1);	/* initially working with
					 * ARG1 */
  Term TermDB;

  ARG3 = Deref(ARG3);
  if (!IsVarTerm(ARG3)) {
    cut_fail();
  }
  DBModule = CurrentModule;
  if (EndOfPAEntr(AtProp = FetchDBPropFromKey(twork, FALSE, "first_instance/3"))) {
    return(FALSE);
  }
#ifdef KEEP_OLD_ENTRIES_HANGING_ABOUT
  ref = AtProp->FirstNEr;
#else
  ref = AtProp->First;
#endif
  while (ref != NIL
	 && (ref->Flags & (DBCode | ErasedMask)))
    ref = NextDBRef(ref);
  if (ref == NIL) {
    cut_fail();
  }
  TRef = MkDBRefTerm(ref);
  /* we have a pointer to the term available */
  if (!(ref->Flags & InUseMask)) {
    ref->Flags |= InUseMask;
    TRAIL_REF(&(ref->Flags));	/* So that fail will erase it */
  }
  while ((TermDB = GetDBTerm(ref)) == (CELL)0) {
    /* oops, we are in trouble, not enough stack space */
    gc(3, ENV, P);
  }
  if (IsVarTerm(TermDB)) {
    unify(TermDB, ARG2);
  } else {
    return(unify(ARG2, TermDB));
  }
  return(unify(ARG3, TRef));
}

/*
 * This is called when we are erasing a data base clause, because we may have
 * pending references 
 */
static void 
ErasePendingRefs(DBRef entryref)
{
  CELL           *cp;
  DBRef           ref;

  if (!(entryref->Flags & DBWithRefs))
    return;
  cp = CellPtr(entryref->Code);
  while ((ref = (DBRef)(*--cp)) != NIL) {
    if ((ref->Flags & DBClMask) && (--(ref->NOfRefsTo) == 0)
	&& (ref->Flags & ErasedMask))
      ErDBE(ref);
  }
}


inline static void 
RemoveDBEntry(DBRef entryref)
{
#ifdef  KEEP_OLD_ENTRIES_HANGING_ABOUT
  DBProp      pp = (DBProp) (entryref->Parent);
#endif

  ErasePendingRefs(entryref);
#ifdef  KEEP_OLD_ENTRIES_HANGING_ABOUT
  if (entryref->Prev == NIL) {
    if (entryref->Next == NIL) {
      pp->First = pp->Last = NIL;
    } else {
      (entryref->Next)->Prev = NIL;
      pp->First = entryref->Next;
    }
  } else if (entryref->Next == NIL) {
    pp->Last = entryref->Prev;
    (entryref->Prev)->Next = NIL;
  }
  else {
    (entryref->Prev)->Next = entryref->Next;
    (entryref->Next)->Prev = entryref->Prev;
  }
#endif
#ifdef DISCONNECT_OLD_ENTRIES
  /* We may be backtracking back to a deleted entry. If we just remove
     the space then the info on the entry may be corrupt.  */
  if ((B->cp_ap == RETRY_C_RECORDED_CODE 
       || B->cp_ap == RETRY_C_DRECORDED_CODE 
       || B->cp_ap == RETRY_C_RECORDEDP_CODE) &&
      EXTRA_CBACK_ARG(3,1) == (CELL)entryref)
    /* make it clear the entry has been released */
    entryref->Flags &= ~InUseMask;
  else
#endif
    {
      FreeDBSpace((char *) entryref);
    }
}

/*
 * Check if the clause is still in his father chain, that might not be true
 * if an abolish had happened after the clause was removed 
 */
inline static int 
StillInChain(CODEADDR cl, PredEntry *pred)
{
  register CODEADDR base, end;

  if (!(pred->PredFlags & DynamicPredFlag))
    return (FALSE);
  base = pred->FirstClause;
  end = pred->LastClause;
  while (cl != base) {
    if (base == end)
      return (FALSE);
    base = NextClause(base);
  }
  return (TRUE);
}


#ifdef DISCONNECT_OLD_ENTRIES

static yamop *
find_next_clause(ref0)
     DBRef ref0;
{
  Register DBRef  ref;
  Int my_age;
  yamop *newp;

  /* fetch ref0 from the instruction we just started executing */
#ifdef DEBUG
  if (!(ref0->Flags & ErasedMask)) {
    Error(SYSTEM_ERROR, TermNil, "find_next_clause (dead clause %x)", ref0);
    return(NIL);
  }
#endif
  my_age = ref0->age;
  /* we were thrown out of the hash chain */
  ref = ((DBProp)(ref0->Parent))->First;
  /* search for an newer entry that is to the left and points to code */
  while (ref != NIL && (ref->age < my_age || !(ref->Flags & DBCode)))
    ref = ref->Next;
  /* no extra alternatives to try, let us leave gracefully */
  if (ref == NIL) {
    return(NIL);
  } else {
    /* OK, we found a clause we can jump to, do a bit of hanky pancking with
       the choice-point, so that it believes we are actually working from that
       clause */
    newp = (yamop *)(ref->Code);
    /* and next let's tell the world this clause is being used, just
       like if we were executing a standard retry_and_mark */
    if (!DynamicFlags(newp) & InUseMask) {
      DynamicFlags(newp) |= InUseMask;
      TRAIL_REF(&DynamicFlags(newp));
    }
    return(newp);
  }
}

/* This procedure is called when a clause is officialy deleted. Its job
   is to find out where the code can go next, if it can go anywhere */
static Int
jump_to_next_dynamic_clause()
{
  DBRef ref = (DBRef)(DBRef)(((yamop *)((CODEADDR)P-(CELL)NEXTOP((yamop *)NIL,sla)))->u.sla.l2);
  yamop *newp = find_next_clause(ref);
  
  if (newp == (yamop *)NULL) {
    cut_fail();
  }
  /* the next alternative to try must be obtained from this clause */
  B->cp_ap = newp;
  /* and next, enter the clause */
  P = NEXTOP(newp,ld);
  /* and return like if nothing had happened. */
  return(TRUE);
}

#endif /* DISCONNECT_OLD_ENTRIES */


static void
MyEraseClause(Clause *clau)
{
  DBRef           ref;
#ifdef KEEP_OLD_ENTRIES_HANGING_ABOUT
  DBRef           next, previous;
  DBProp          father;
  PredEntry      *pred;
  unsigned int    arity;
  Atom            name;
#endif
  SMALLUNSGN      clmask = clau->ClFlags;

  if (clmask & InUseMask)
    return;
#ifdef KEEP_OLD_ENTRIES_HANGING_ABOUT
  /* after the fail we have a DBRef */
  ref = (DBRef) NEXTOP(clau->ClCode,ld)->u.d.d;
  if (ref->Flags & InUseMask)
    return;
  next = ref->Next;
  previous = ref->Prev;
  while (next != NIL && next->Flags & DBNoCode)
    next = next->Next;
  while (previous != NIL && previous->Flags & DBNoCode)
    previous = previous->Prev;
  if (previous != NIL && next != NIL) {
    yamop  *previous_code = (yamop *)previous->Code;

    previous_code->u.ld.d = next->Code;
  } else {
    father = (DBProp) (ref->Parent);
    if ((arity = father->ArityOfDB) == 0)
      name = (Atom) father->FunctorOfDB;
    else
      name = NameOfFunctor(father->FunctorOfDB);
    DBModule = father->ModuleOfDB;
    pred = RepPredProp(PredProp(name, arity));
    if (StillInChain((CODEADDR)(clau->ClCode), pred)) {
      if (previous == NIL && next != NIL) {
	CODEADDR        second;
	yamop *last;
	second = NextClause(pred->FirstClause);
	pred->FirstClause = second;
	last = (yamop *)(pred->CodeOfPred);
	last->u.ld.d = second;
      } else if (previous != NIL) {
	yamop *previousoflast = (yamop *)(previous->Code);
	pred->LastClause = (CODEADDR)previousoflast;
	previousoflast->u.ld.d = pred->CodeOfPred;
      } else {
	FreeCodeSpace(((char *) ClauseCodeToClause(pred->CodeOfPred)));
	pred->LastClause = pred->FirstClause = NIL;
	p->OpcodeOfPred = opcode(_undef_p);
	p->TrueCodeOfPred = p->CodeOfPred =
	  (CODEADDR)(&(p->OpcodeOfPred)); 
      }
    }
  }
  if ( P == clau->ClCode) {
    yamop   *nextto;
    P = (yamop *)(RTRYCODE);
    nextto = (yamop *)RTRYCODE;
    nextto->u.ld.d = clau->ClCode->u.ld.d;
    nextto->u.ld.s = clau->ClCode->u.ld.s;
    nextto->u.ld.p = clau->ClCode->u.ld.p;
  }
#endif /* KEEP_OLD_ENTRIES_HANGING_ABOUT */
#ifdef DISCONNECT_OLD_ENTRIES
  ref = (DBRef) NEXTOP(clau->ClCode,ld)->u.sla.l2;
  /* don't do nothing if the reference is still in use */
  if (ref->Flags & InUseMask)
    return;
  if ( P == clau->ClCode ) {
    yamop *np = (yamop *)RTRYCODE;
    /* make it the next alternative */
    np->u.ld.d = (CODEADDR)find_next_clause((DBRef)(NEXTOP(P,ld)->u.sla.l2));
    if (np->u.ld.d == NULL)
      P = (yamop *)FAILCODE;
    else {
      /* with same arity as before */
      np->u.ld.s = P->u.ld.s;
      np->u.ld.p = P->u.ld.p;
      /* go ahead and try this code */
      P = np;
    }
  } else {
#endif /* DISCONNECT_OLD_ENTRIES */
    FreeCodeSpace((char *)clau);
#ifdef DEBUG
    if (ref->NOfRefsTo)
      fprintf(stderr, "Error: references to dynamic clause\n");
#endif
    RemoveDBEntry(ref);
#if DISCONNECT_OLD_ENTRIES
  }
#endif
}

void 
ErCl(CODEADDR clau)
{
  YAPEnterCriticalSection();
  MyEraseClause(ClauseFlagsToClause(clau));
  YAPLeaveCriticalSection();
}


static void 
PrepareToEraseClause(Clause *clau, DBRef dbr)
{  yamop          *code_p;

  /* no need to erase what has been erased */
  if (clau->ClFlags & ErasedMask)
    return;
  clau->ClFlags |= ErasedMask;
  /* skip mask */
  code_p = clau->ClCode;
  /* skip retry instruction */
#ifdef KEEP_OLD_ENTRIES_HANGING_ABOUT
  code_p = NEXTOP(code_p, ld);
  /* in this case, a fail will send you back to the next clause */
  code_p->opc = opcode(_op_fail);
  code_p->u.d.d = (CODEADDR)(dbr);
#endif
#ifdef DISCONNECT_OLD_ENTRIES
  /* we can remove the entry from the list of alternatives for the
     goal immediately */
  {
    DBProp father;
    Int arity;
    Atom name;
    Int DBModule;
    PredEntry *pred;
    /* first we get the next clause */
    CODEADDR next = code_p->u.ld.d;
    /* then we get the previous clause */
    CODEADDR previous =  (CODEADDR)(clau->u.ClPrevious);
    CODEADDR clau_code;

    /* next we check if we still have clauses left in the chain */
    if (previous != next) {
      yamop  *previous_code = (yamop *)previous;
      Clause *next_cl = ClauseCodeToClause(next);
      /* we do, let's say the previous now backtracks to the next */
      previous_code->u.ld.d = next;
      /* and tell next who it is the previous element */
      next_cl->u.ClPrevious = previous_code;
    }
    /* that's it about setting up the code, now let's tell the
       predicate entry that a clause left. */
    father = (DBProp) dbr->Parent;
    /* inefficient, but that will do for the moment, sir. */
    if ((arity = father->ArityOfDB) == 0)
      name = (Atom) father->FunctorOfDB;
    else
      name = NameOfFunctor(father->FunctorOfDB);
    DBModule = father->ModuleOfDB;
    pred = RepPredProp(PredProp(name, arity));
    /* got my pred entry, let's have some fun! */
    clau_code = (CODEADDR)(clau->ClCode);
    if (pred->FirstClause == pred->LastClause) {
#ifdef DEBUG
      if (pred->FirstClause != clau_code) {
	/* sanity check */
	Error(SYSTEM_ERROR, TermNil, "Prepare to erase clause for %s/%d",RepAtom(name)->StrOfAE,arity);
	return;
      }
#endif
      /* nothing left here, let's clean the shop */
      FreeCodeSpace(((char *) ClauseCodeToClause(pred->CodeOfPred)));
      pred->LastClause = pred->FirstClause = NIL;
      pred->OpcodeOfPred = opcode(_undef_p);
      pred->TrueCodeOfPred = pred->CodeOfPred =
	(CODEADDR)(&(pred->OpcodeOfPred)); 
    } else if (clau_code == pred->FirstClause) {
      pred->FirstClause = next;
    } else if (clau_code == pred->LastClause) {
      pred->LastClause = previous;
    }
  }
  /* make sure we don't directly point to anyone else */
  code_p->u.ld.d = (CODEADDR)code_p;
  /* now, put some code so that backtracks to here will survive */
  code_p = NEXTOP(code_p, ld);
  /* in this case, a failed clause should go to the data base and find
     out  what is the next clause, if there is one */
  code_p->opc = opcode(_call_cpred);
  code_p->u.sla.l = (CODEADDR)(&jump_to_next_dynamic_clause);
  code_p->u.sla.l2 = (CODEADDR)(dbr);
#endif /* DISCONNECT_OLD_ENTRIES */
}

void 
ErDBE(DBRef entryref)
{
  YAPEnterCriticalSection();
  if ((entryref->Flags & DBCode) && entryref->Code) {
    Clause *clau = ClauseCodeToClause(entryref->Code);
    if (clau->ClFlags & InUseMask || entryref->NOfRefsTo != 0) {
      PrepareToEraseClause(clau, entryref);
    } else {
      if (!(clau->ClFlags & ErasedMask))
	PrepareToEraseClause(clau, entryref);
      MyEraseClause(clau);
    }
  } else if (!(entryref->Flags & InUseMask)) {
    if (entryref->NOfRefsTo == 0) 
      RemoveDBEntry(entryref);
    else if (!(entryref->Flags & ErasedMask)) {
      /* oops, I cannot remove it, but I at least have to tell
	 the world what's going on */
      entryref->Flags |= ErasedMask;
#ifdef DISCONNECT_OLD_ENTRIES
      entryref->Next = entryref->Prev = NIL;
#endif
    }
  }
  YAPLeaveCriticalSection();
}

static void
EraseEntry(DBRef entryref)
{
  DBProp          p;

  if (entryref->Flags & ErasedMask)
    return;
  entryref->Flags |= ErasedMask;
  /* update FirstNEr */
  p = (DBProp)(entryref->Parent);
#ifdef KEEP_OLD_ENTRIES_HANGING_ABOUT
  if (p->FirstNEr == entryref) {
    DBRef           q = entryref->Next;
    while (q != NIL && (q->Flags & ErasedMask))
      q = q->Next;
    p->FirstNEr = q;
  }
#endif /* KEEP_OLD_ENTRIES_HANGING_ABOUT */
#ifdef DISCONNECT_OLD_ENTRIES
  /* exit the db chain */
  if (entryref->Next != NIL) {
    entryref->Next->Prev = entryref->Prev;
  } else {
    p->Last = entryref->Prev;
  }
  if (entryref->Prev != NIL)
    entryref->Prev->Next = entryref->Next;
  else
    p->First = entryref->Next;
  /* make sure we know the entry has been removed from the list */
  entryref->Next = NIL;
#endif /* DISCONNECT_OLD_ENTRIES */
  if (!(entryref->Flags & InUseMask)) {
    ErDBE(entryref);
  } else if ((entryref->Flags & DBCode) && entryref->Code) {
    PrepareToEraseClause(ClauseCodeToClause(entryref->Code), entryref);
  }
}

/* erase(+Ref)	 */
static Int 
p_erase(void)
{
  Term t1 = Deref(ARG1);

  if (IsVarTerm(t1) || !IsDBRefTerm(t1))
    return (FALSE);
  EraseEntry(DBRefOfTerm(t1));
  return (TRUE);
}

/* eraseall(+Key)	 */
static Int 
p_eraseall(void)
{
  Register Term   twork = Deref(ARG1);
  Register DBRef  entryref;
  DBProp          p;

  DBModule = 0;
  if (EndOfPAEntr(p = FetchDBPropFromKey(twork, FALSE, "eraseall/3"))) {
    return(TRUE);
  }
  entryref = FrstDBRef(p);
  do {
    DBRef next_entryref;

    while (entryref != NIL &&
	   (entryref->Flags & (DBCode | ErasedMask)))
      entryref = NextDBRef(entryref);
    if (entryref == NIL)
      break;
    next_entryref = NextDBRef(entryref);
#ifdef DISCONNECT_OLD_ENTRIES
    /* exit the db chain */
    if (entryref->Next != NIL) {
      entryref->Next->Prev = entryref->Prev;
    } else {
      p->Last = entryref->Prev;
    }
    if (entryref->Prev != NIL)
      entryref->Prev->Next = entryref->Next;
    else
      p->First = entryref->Next;
    /* make sure we know the entry has been removed from the list */
    entryref->Next = entryref->Prev = NIL;
#endif
    if (!(entryref->Flags & InUseMask))
      ErDBE(entryref);
    else {
      entryref->Flags |= ErasedMask;
    }
    entryref = next_entryref;
  } while (entryref != NIL);
#ifdef KEEP_OLD_ENTRIES_HANGING_ABOUT
  RepDBProp(AtProp)->FirstNEr = NIL;
#endif
  return (TRUE);
}


/* erased(+Ref) */
static Int 
p_erased(void)
{
  Term            t = Deref(ARG1);

  if (IsVarTerm(t) || !IsDBRefTerm(t))
    return (FALSE);
  return (DBRefOfTerm(t)->Flags & ErasedMask);
}


/* instance(+Ref,?Term) */
static Int 
p_instance(void)
{
  Term            TermDB;
  Term t1 = Deref(ARG1);

  if (IsVarTerm(t1) || !IsDBRefTerm(t1))
    return (FALSE);
  while ((TermDB = GetDBTerm(DBRefOfTerm(t1))) == (CELL)0) {
    /* oops, we are in trouble, not enough stack space */
    gc(2, ENV, P);
    t1 = Deref(ARG1);
  }
  return (unify(ARG2, TermDB));
}

inline static int 
NotActiveDB(DBRef my_dbref)
{
  while (my_dbref && (my_dbref->Flags & (DBCode | ErasedMask)))
    my_dbref = my_dbref->Next;
  return (my_dbref == NIL);
}

inline static DBEntry *
NextDBProp(PropEntry *pp)
{
  while (!EndOfPAEntr(pp) && ((pp->KindOfPE != DBProperty) ||
			      NotActiveDB(((DBProp) pp)->First)))
    pp = RepProp(pp->NextOfPE);
  return ((DBEntry *)pp);
}

static Int 
init_current_key(void)
{				/* current_key(+Atom,?key)	 */
  Int             i = 0;
  DBEntry        *pp;
  Atom            a;
  Term t1 = ARG1;

  t1 = Deref(ARG1);
  if (!IsVarTerm(t1)) {
    if (IsAtomTerm(t1))
      a = AtomOfTerm(t1);
    else {
      cut_fail();
    }
  } else
    while ((a = HashChain[i]) == NIL)
      i++;
  pp = NextDBProp(RepProp(RepAtom(a)->PropOfAE));
  EXTRA_CBACK_ARG(2,3) = MkAtomTerm(a);
  EXTRA_CBACK_ARG(2,2) = MkIntTerm(i);
  EXTRA_CBACK_ARG(2,1) = (CELL)pp;
  return (cont_current_key());
}

static Int 
cont_current_key(void)
{
  unsigned int    arity;
  Functor         functor;
  Term            term, AtT;
  CELL		 *p = H;
  Atom            a = AtomOfTerm(EXTRA_CBACK_ARG(2,3));
  Int             i = IntOfTerm(EXTRA_CBACK_ARG(2,2));
  Term            first = Deref(ARG1);
  DBEntry        *pp = (DBEntry *) EXTRA_CBACK_ARG(2,1);

  if (EndOfPAEntr(pp) && IsAtomTerm(first)) {
    cut_fail();
  }
  while (EndOfPAEntr(pp)) {
    if ((a = RepAtom(a)->NextOfAE) == NIL) {
      do
	i++;
      while (i < MaxHash && (a = HashChain[i]) == NIL);
      if (i == MaxHash) {
	cut_fail();
      }
      EXTRA_CBACK_ARG(2,2) = MkIntTerm(i);
    }
    if (!EndOfPAEntr(pp = NextDBProp(RepProp(RepAtom(a)->PropOfAE))))
      EXTRA_CBACK_ARG(2,3)  = (CELL) MkAtomTerm(a);
  }
  EXTRA_CBACK_ARG(2,1) = (CELL)(NextDBProp(RepProp(pp->NextOfPE)));
  arity = (unsigned int)(pp->ArityOfDB);
  if (arity == 0) {
    term = AtT = MkAtomTerm(a);
  } else {
    unsigned int j;

    for (j = 0; j < arity; j++) {
      /* p[j] = */ (void)MkVarTerm();
    }
    functor = MkFunctor(a, arity);
    term = MkApplTerm(functor, arity, p);
    AtT = MkAtomTerm(a);
  }
  return (unify_constant(ARG1, AtT) && unify(ARG2, term));
}

static Int 
p_init_queue(void)
{
  db_queue *dbq;
  Term t;

  while ((dbq = (db_queue *)AllocDBSpace(sizeof(db_queue))) == NIL) {
    if (!growheap()) {
      Abort("[ SYSTEM ERROR: YAP failed to reserve space in growheap ]\n");
      return(FALSE);
    }
  }
  dbq->id = FunctorDBRef;
  dbq->Flags = DBClMask;
  dbq->FirstInQueue = dbq->LastInQueue = NIL;
  t = MkDBRefTerm((DBRef)dbq);
  return(unify(ARG1, t));
}

static Int 
p_enqueue(void)
{
  Term Father = Deref(ARG1);
  DBRef x;
  db_queue *father_key;

  if (IsVarTerm(Father)) {
    Error(INSTANTIATION_ERROR, Father, "enqueue");
    return(FALSE);
  } else if (!IsDBRefTerm(Father)) {
    Error(TYPE_ERROR_DBREF, Father, "enqueue");
    return(FALSE);
  } else
    father_key = (db_queue *)DBRefOfTerm(Father);
  while ((x = CreateDBStruct(Deref(ARG2), (DBProp)NIL,
			  InQueue)) == NULL) {
    switch(DBErrorFlag) {
    case NO_ERROR_IN_DB:
#ifdef DEBUG
      Abort("[ SYSTEM ERROR: no error but null return in enqueue/2 ]\n");
#endif
      break;
    case SOVF_ERROR_IN_DB:
      if (!gc(2, ENV, P)) {
	Abort("[ SYSTEM ERROR: YAP could not grow stack in enqueue/2 ]\n");
	return(FALSE);
      } else
	break;
    case OVF_ERROR_IN_DB:
      if (!growheap()) {
	Abort("[ SYSTEM ERROR: YAP could not grow heap in enqueue/2 ]\n");
	return(FALSE);
      } else
	break;
    default:
      Error(DBErrorNumber, DBErrorTerm, DBErrorMsg);
      return(FALSE);
    }
  }
  if (father_key->LastInQueue != NIL)
    father_key->LastInQueue->Parent = ParentCast(x);
  father_key->LastInQueue = x;
  if (father_key->FirstInQueue == NIL)
    father_key->FirstInQueue = x;
  return(TRUE);
}

/* when reading an entry in the data base we are making it accessible from
   the outside. If the entry was removed, and this was the last pointer, the
   target entry would be immediately removed, leading to dangling pointers.
   We avoid this problem by making every entry accessible. 

   Note that this could not happen with recorded, because the original db
   entry itself is still accessible from a trail entry, so we could not remove
   the target entry,
 */
static void
keepdbrefs(DBRef entryref)
{
  CELL           *cp;
  DBRef           ref;

  if (!(entryref->Flags & DBWithRefs))
    return;
  cp = CellPtr(entryref->Code);
  while ((ref = (DBRef)(*--cp)) != NIL) {
    if(!(ref->Flags & InUseMask)) {
      ref->Flags |= InUseMask;
      TRAIL_REF(&(ref->Flags));	/* So that fail will erase it */
    }
  }

}

static Int 
p_dequeue(void)
{
  db_queue *father_key;
  DBRef cur_instance;
  Term Father = Deref(ARG1);

  if (IsVarTerm(Father)) {
    Error(INSTANTIATION_ERROR, Father, "dequeue");
    return(FALSE);
  } else if (!IsDBRefTerm(Father)) {
    Error(TYPE_ERROR_DBREF, Father, "dequeue");
    return(FALSE);
  } else
    father_key = (db_queue *)DBRefOfTerm(Father);
  if ((cur_instance = father_key->FirstInQueue) == NIL) {
    /* an empty queue automatically goes away */
    FreeDBSpace((char *) father_key);
    return(FALSE);
  } else {
    Term TDB;
    while ((TDB = GetDBTerm(father_key->FirstInQueue)) == (CELL)0) {
      /* oops, we are in trouble, not enough stack space */
      gc(2, ENV, P);
    }
    if (cur_instance == father_key->LastInQueue)
      father_key->FirstInQueue = father_key->LastInQueue = NIL;
    else father_key->FirstInQueue = (DBRef)(cur_instance->Parent);
    /* release space for cur_instance */
    keepdbrefs(cur_instance);
    ErasePendingRefs(cur_instance);
    FreeDBSpace((char *) cur_instance);
    return(unify(ARG2, TDB));
  }
}

/*

  This is a hack, to steal the first element of a key.

   It first fetches the first element in the chain, and then erases it
   through its reference.

  Be careful when using this routine. It is especially evil because if
  the term is ground it should be copied to the stack, as space for
  the entry may be deleted. For the moment, the terms I want are just
  integers, so no problemo, amigo.

 */
static Int
StealFirstFromDB(Atom key, Int arity)
{
  Prop            AtProp;
  Register DBRef  ref;
  Term            TermDB;
  Int out;

  /* get the DB property */
  if ((AtProp = FindDBProp(key, arity)) == NIL) {
    return(TermNil);
  }
  /* get the first entry */
#ifdef KEEP_OLD_ENTRIES_HANGING_ABOUT
  ref = RepDBProp(AtProp)->FirstNEr;
#else
  ref = RepDBProp(AtProp)->First;
#endif
  /* is there anyone home ? */
  while (ref != NIL
	 && (ref->Flags & (DBCode | ErasedMask)))
    ref = NextDBRef(ref);
  if (ref == NIL) {
    return(TermNil);
  }
  /* get our fine  term */
  if ((TermDB = GetDBTerm(ref)) == (CELL)0) {
    /* oops, we are in trouble, not enough stack space */
    return(0);
  }
  if (IsVarTerm(TermDB) || !IsApplTerm(TermDB))
    /* it's not a wonderful world afterall */
    return(0);
  out = IntOfTerm(ArgOfTerm(1,TermDB));
  /* next, make it disappear from the DB */
  EraseEntry(ref);
  /* now, return what once was there, only nevermore */
  return(out);
}

Int
SetDBForThrow(Term Message)
{
  Int cut_pt;

  /* who's gonna catch us? */
  DBModule = 0;
  if ((cut_pt = StealFirstFromDB(AtomCatch, 0)) == 0) {
    /* ooops, babe we are in trouble */
    return(0L);
  }
  /* OK, we've got the place to cut to, next store the new throw */
  if (record(MkFirst, MkAtomTerm(AtomThrow), Message, TermNil) == NIL)
    return (0L);
  else
    /* off we go, to see the wizard of Oz */
    return(cut_pt);  
}

void 
InitBackDB(void)
{
  InitCPredBack("recorded", 3, 3, in_rded, co_rded, SafePredFlag|SyncPredFlag);
  /* internal version, just to prevent the debugger from nosying around */
  RETRY_C_RECORDED_CODE = NEXTOP((yamop *)
    (RepPredProp(PredProp(LookupAtom("recorded"), 3))->FirstClause),lds);
  InitCPredBack("$recorded", 3, 3, in_rded, co_rded, SafePredFlag|SyncPredFlag);
  RETRY_C_DRECORDED_CODE = NEXTOP((yamop *)
    (RepPredProp(PredProp(LookupAtom("$recorded"), 3))->FirstClause),lds);
  InitCPredBack("$recordedp", 3, 3, in_rdedp, co_rdedp, SafePredFlag|SyncPredFlag);
  RETRY_C_RECORDEDP_CODE = NEXTOP((yamop *)
    (RepPredProp(PredProp(LookupAtom("$recordedp"), 3))->FirstClause),lds);
  InitCPredBack("current_key", 2, 4, init_current_key, cont_current_key,
		SafePredFlag|SyncPredFlag);
}

/* given a key, find the clock number for the first entry */
/* $db_key_to_nb(+Key,-Int) */ 
static Int
p_first_age(void)
{
  Term t1 = Deref(ARG1);
  Term to;
  DBProp AtProp;

  DBModule = CurrentModule;
  if (EndOfPAEntr(AtProp = FetchDBPropFromKey(t1, FALSE, "first_age/3"))) {
    return(FALSE);
  }
#ifdef KEEP_OLD_ENTRIES_HANGING_ABOUT
  if (AtProp->FirstNEr == NULL)
    return(FALSE);
  to = MkIntegerTerm(AtProp->FirstNEr->age);
#else
  if (AtProp->First == NULL)
    to = MkIntegerTerm(AtProp->age);
  else
    to = MkIntegerTerm(AtProp->First->age);
#endif
  return(unify(ARG2,to));
}

/* given an integer, and a reference to the fist element, find the
   corresponding reference, if one exists. */
/* $db_key_to_nb(+Age,+Key,+Flags,-Ref) */ 
static Int
p_db_nb_to_ref(void)
{
  Term t1 = Deref(ARG1);
  Term t2 = Deref(ARG2);
  Term t3 = Deref(ARG3);
  Term tref;
  DBRef myref;
  Int age, flagsToAvoid;
  DBProp AtProp;

  if (IsVarTerm(t1))
    return(FALSE);
  if (IsIntTerm(t1))
    age = IntOfTerm(t1);
  else if (IsLongIntTerm(t1))
    age = LongIntOfTerm(t1);
  else return(FALSE);
  DBModule = CurrentModule;
  if (EndOfPAEntr(AtProp = FetchDBPropFromKey(t2, FALSE, "recorded/3"))) {
    return(FALSE);
  }
#ifdef KEEP_OLD_ENTRIES_HANGING_ABOUT
  myref = AtProp->FirstNEr;
#else
  myref = AtProp->First;
#endif
  if (IsVarTerm(t3))
    return(FALSE);
  if (IsIntTerm(t3))
    flagsToAvoid = IntOfTerm(t3);
  else if (IsLongIntTerm(t3))
    flagsToAvoid = LongIntOfTerm(t3);
  else return(FALSE);
  while (myref != NIL
	 && ((myref->Flags & flagsToAvoid)
	     ||  myref->age < age))
    myref = NextDBRef(myref);
  if (myref == NIL || myref->age != age) {
    return(FALSE);
  }
  tref = MkDBRefTerm(myref);
  if (!(myref->Flags & InUseMask)) {
    myref->Flags |= InUseMask;
    TRAIL_REF(&(myref->Flags));	/* So that fail will erase it */
  }
  return(unify(ARG4,tref));
}

/* given a key, find the clock number for the last entry */
/* $db_last_age(+Key,-Int) */ 
static Int
p_last_age(void)
{
  Term t1 = Deref(ARG1);
  DBProp AtProp;
  Term last_age;

  DBModule = CurrentModule;
  if ((AtProp = FetchDBPropFromKey(t1, FALSE, "$last_age/2")) == NIL) {
    return(FALSE);
  }
  last_age = MkIntegerTerm(AtProp->age);
  return(unify(ARG2,last_age));
}


void 
InitDBPreds(void)
{
  InitCPred("$recorda", 3, p_rcda, SafePredFlag|SyncPredFlag);
  InitCPred("$recordz", 3, p_rcdz, SafePredFlag|SyncPredFlag);
  InitCPred("$recordap", 3, p_rcdap, SafePredFlag|SyncPredFlag);
  InitCPred("$recordzp", 3, p_rcdzp, SafePredFlag|SyncPredFlag);
  InitCPred("$recordap", 4, p_drcdap, SafePredFlag|SyncPredFlag);
  InitCPred("$recordzp", 4, p_drcdzp, SafePredFlag|SyncPredFlag);
  InitCPred("$recordaifnot", 3, p_rcdaifnot, SafePredFlag|SyncPredFlag);
  InitCPred("$recordzifnot", 3, p_rcdzifnot, SafePredFlag|SyncPredFlag);
  InitCPred("erase", 1, p_erase, SafePredFlag|SyncPredFlag);
  InitCPred("erased", 1, p_erased, TestPredFlag | SafePredFlag|SyncPredFlag);
  InitCPred("instance", 2, p_instance, SyncPredFlag);
  InitCPred("eraseall", 1, p_eraseall, SafePredFlag|SyncPredFlag);
  InitCPred("$record_stat_source", 3, p_rcdstatp, SafePredFlag|SyncPredFlag);
  InitCPred("$some_recordedp", 1, p_somercdedp, SafePredFlag|SyncPredFlag);
  InitCPred("$first_instance", 3, p_first_instance, SafePredFlag|SyncPredFlag);
  InitCPred("$init_db_queue", 1, p_init_queue, SafePredFlag|SyncPredFlag);
  InitCPred("$db_enqueue", 2, p_enqueue, SyncPredFlag);
  InitCPred("$db_dequeue", 2, p_dequeue, SyncPredFlag);
  InitCPred("$db_first_age", 2, p_first_age, TestPredFlag|SafePredFlag|SyncPredFlag);
  InitCPred("$db_nb_to_ref", 4, p_db_nb_to_ref, TestPredFlag|SafePredFlag);
  InitCPred("$db_last_age", 2, p_last_age, TestPredFlag|SafePredFlag|SyncPredFlag);
}
